// Copyright (C) 2006 Pushok Software http://www.pushok.com/soft_svncom.php
//
// LICENSE
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License (GPL)
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// To read the license please visit http://www.gnu.org/copyleft/gpl.html
//
// Subversion is Copyrght of(c) 2000-2004 CollabNet.  All rights reserved.
// Full license available here http://subversion.tigris.org/license-1.html.

#include "stdafx.h"
#include "SVNClient.h"

#include "client.h"
#include "svn_config.h"
#include "libsvn_wc\wc.h"
#include "libsvn_wc\entries.h"

#include "Helpers.h"
#include "SVNWcStatus.h"
#include "SVNWcNotify.h"
#include "SVNWcLog.h"
#include "SVNWcInfo.h"
#include "SVNWcPropList.h"
#include "SVNWcAnn.h"

STDMETHODIMP CSVNClient::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_ISVNClient,
		&IID_ISVNClient2,
		&IID_ISVNClient3,
		&IID_ISVNClient4,
		&IID_ISVNClient5
	};

	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

#define BEGIN_SVN_CALL \
	HRESULT hr=S_OK;\
	{\
	Lock();\
	hr = InitClient(); if (FAILED(hr)) return hr;\
	m_bCancelRequested = false;\
	hr = S_OK;\
	apr_pool_t *pFuncPool = svn_pool_create(NULL);\
	try\
	{

#define END_SVN_CALL \
	}\
	catch (...)\
	{\
	ATLASSERT(FALSE);\
	hr = E_UNEXPECTED;\
	TRACELOG_CHAR("Exception is caught")\
	TRACELOG_NEWLINE\
	}\
	goto end_svn_call_;\
	end_svn_call:\
	TRACELOG_END("Exception is caught 2")\
	end_svn_call_:\
	svn_pool_destroy(pFuncPool);\
	Unlock();\
	TRACELOG_END("End Svn Call")\
	}

#define SVN_CALL_ERROR(_hr,msg) \
	{\
		hr = _hr;\
		hr = AtlReportError(__uuidof(SVNClient),CString(msg),__uuidof(ISVNClient),hr);\
		TRACELOG_CHAR("Generate exception")\
		TRACELOG_NEWLINE\
		TRACELOG_CHAR((char *)msg[0])\
		TRACELOG_NEWLINE\
		goto end_svn_call;\
	}

#define SVN_CALL_SVNERROR(err) \
	{\
		svn_error_t *_err = err;\
		if (_err)\
		{\
			CString sMessage = SvnErrorToString(_err);\
			svn_error_clear(_err);\
			hr = AtlReportError(__uuidof(SVNClient),Utf82Ole(sMessage),__uuidof(ISVNClient),hr);\
			TRACELOG_CHAR("Generate exception")\
			TRACELOG_NEWLINE\
			TRACELOG_CHAR((LPCSTR)sMessage[0])\
			TRACELOG_NEWLINE\
			goto end_svn_call;\
		};\
	}

// Decode error messages like the native SVN client does
static CString SvnErrorToString(svn_error_t *err)
{
	CString sMessage;
	std::set<apr_status_t> empties;
	while (err)
	{
		if (err->message || !empties.count(err->apr_err))
		{
			// Decode and print error message
			char errbuf[256];
			const char *err_string = NULL;
			if (err->message)
				err_string = err->message;
			else
			{
				// Is this a SVN-specific error code?
				if ((err->apr_err > APR_OS_START_USEERR) && (err->apr_err <= APR_OS_START_CANONERR))
					err_string = svn_strerror(err->apr_err, errbuf, sizeof(errbuf));
				// Otherwise, this must be an APR error code
				else
				{
					svn_error_t *temp_err = svn_utf_cstring_to_utf8(&err_string,
							apr_strerror(err->apr_err, errbuf, sizeof(errbuf)), err->pool);
					if (temp_err)
					{
						svn_error_clear(temp_err);
						err_string = "Can't recode error string from APR";
					}
				}
				// Remember the code of empty error, so it will be printed only once
				empties.insert(err->apr_err);
			}
			if (!sMessage.IsEmpty())
				sMessage += '\n';
			ATLASSERT(err_string != NULL);
			sMessage += err_string;
		}
		err = err->child;
	}
	return sMessage;
}

void client_progress_func(apr_off_t progress, apr_off_t total, void *_baton, apr_pool_t *pool)
{
	// safely obtain pointer to client class
	if (!_baton) return;
	_ISVNClient *oClient = (_ISVNClient *)_baton;

	while (progress>0xffffffff) progress/=10;
	while (total>0xffffffff) total/=10;

	// fire event outside
	oClient->FireWcProgress(static_cast<long>(progress), static_cast<long>(total));
}

// the notification function, forward call as events to COM client
void client_notify_func(void *_baton, const svn_wc_notify_t *msg, apr_pool_t *pool)
{
	// safely obtain pointer to client class
	if (!_baton) return;
	CSVNClient::ClientNotifyBaton  *baton = (CSVNClient::ClientNotifyBaton*)_baton;
	_ISVNClient *oClient = baton->oClient;
	ISVNWcNotifyVector *oNotifyVector = baton->oNotifyVector;

	// create notify action
	CComPtr<ISVNWcNotify> oNotify;
	oNotify.CoCreateInstance(__uuidof(SVNWcNotify));

	oNotify->put_Path(Utf82Ole(msg->path));
	oNotify->put_Action((SvnWcNotifyAction)msg->action);
	oNotify->put_Kind((SvnNodeKind)msg->kind);
	if (msg->err)
	{
		oNotify->put_CodeError(msg->err->apr_err);
		oNotify->put_StrError(Utf82Ole(msg->err->message));
	}
	else
		oNotify->put_CodeError(0);

	oNotify->put_ContState((SvnWcNotifyState)msg->content_state);
	oNotify->put_PropState((SvnWcNotifyState)msg->prop_state);
	oNotify->put_Revision(msg->revision);

	// fire event outside
	oClient->FireWcNotify(oNotify);

	if (oNotifyVector)
		oNotifyVector->Add(oNotify);
}

svn_error_t *client_get_log_message (const char **log_msg, const char **tmp_file, 
									  apr_array_header_t *commit_items, void *baton, apr_pool_t *pool)
{
    *log_msg = (const char*)baton;
	*tmp_file = NULL;
    return SVN_NO_ERROR;
}

// the cancel notification function
svn_error_t *client_check_cancel (void *baton)
{
	// safely obtain pointer to client class
	if (!baton) return SVN_NO_ERROR;
	_ISVNClient *oClient=(_ISVNClient *)baton;

	// well done, got it, do what we need, it is safe now

	oClient->FireIdle();
	LONG bValue = false;
	oClient->IsCancelRequested(&bValue);

	if (bValue)
		return svn_error_create (SVN_ERR_CANCELLED, NULL, "Client request cancel");
	else
		return SVN_NO_ERROR;
}

/* Version compatibility check */
static svn_error_t *check_lib_versions (void)
{
	static const svn_version_checklist_t checklist[] =
	{
		{ "svn_subr",   svn_subr_version },
		{ "svn_client", svn_client_version },
		{ "svn_wc",     svn_wc_version },
		{ "svn_ra",     svn_ra_version },
		{ "svn_delta",  svn_delta_version },
		{ NULL, NULL }
	};

	SVN_VERSION_DEFINE (my_version);
	return svn_ver_check_list (&my_version, checklist);
}

STDMETHODIMP CSVNClient::InitClient(void)
{
	TRACELOG_BEGIN("InitClient")
	// check are we initialized or no
	if (m_bInitialized)
	{
		TRACELOG_END("InitClient already started.")
		return S_OK;
	}
	
	// not, let's do it
	apr_status_t status;
	svn_error_t *err;

	status = apr_initialize();
	if (status)
	{
		TRACELOG_END("InitClient call exception.")
		return AtlReportError(__uuidof(SVNClient),_T("Cannot initialize apr pool"));
	}

	// init client pool
	m_pPool = svn_pool_create(NULL);

	/* Check library versions */
	err = check_lib_versions ();
	if (err)
	{
		TRACELOG_END("InitClient call exception.")
		return AtlReportError(__uuidof(SVNClient),_T("Cannot initialize apr pool"));
	}

	/* Initialize the RA library. */
	err = svn_ra_initialize (m_pPool);
	if (err)
	{
		TRACELOG_END("InitClient call exception.")
		return AtlReportError(__uuidof(SVNClient),_T("Cannot initialize apr pool"));
	}

	/* Create a pool for use by the UTF-8 routines.  It will be cleaned
    up by APR at exit time. */
	svn_utf_initialize (svn_pool_create (NULL));


	// init the svn content
	svn_client_create_context (&m_pCtx, m_pPool);
	m_pCtx->notify_func2 = &client_notify_func;
	m_oClientNotifyBaton.oClient = this;
	m_oClientNotifyBaton.oNotifyVector = NULL;
	m_pCtx->notify_baton2 = &m_oClientNotifyBaton;
	m_pCtx->cancel_func = client_check_cancel;
	m_pCtx->cancel_baton = (_ISVNClient *)this;

	m_pCtx->log_msg_func = &client_get_log_message;
	m_pCtx->progress_baton = (_ISVNClient *)this; 
	m_pCtx->progress_func = &client_progress_func;

	// load config
	if (m_sOptConfigDir.GetLength())
	{
		svn_config_ensure (Ansi2Utf8(m_sOptConfigDir), m_pPool);
		svn_config_get_config (&(m_pCtx->config),Ansi2Utf8(m_sOptConfigDir),m_pPool);
	}
	else
	{
		svn_config_ensure (NULL, m_pPool);
		svn_config_get_config (&(m_pCtx->config),NULL,m_pPool);
	}

	// dummy auth setup
	_SetupAuth(NULL, &m_dwAuthCookie);
	m_bAuthExternal = false;
	m_bInitialized = true;

	TRACELOG_END("InitClient")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(S_OK)	TRACELOG_NEWLINE

	return S_OK;
}

STDMETHODIMP CSVNClient::_SetupAuth(LONG	nCookie, LONG * ppCookie	)
{
	TRACELOG_BEGIN("_SetupAuth")

	TRACELOG_TITLE("nCookie: ")	TRACELOG_NUMBER(nCookie)	TRACELOG_NEWLINE
	TRACELOG_TITLE("ppCookie: ")	TRACELOG_NUMBER(*ppCookie)	TRACELOG_NEWLINE

	CComPtr<ISVNAuthManager> oAuth;
	HRESULT  hr = oAuth.CoCreateInstance(__uuidof(SVNAuthManager),NULL, CLSCTX_INPROC_SERVER);

	DWORD dwOldCookie(m_dwAuthCookie);

	CString sErr = "";
	CComPtr<ISVNAuthProvider> oAuthProvider;
	// create or advise to new
	if (nCookie==NULL)
	{
		// use latest if it is internal or create new
		if (m_bAuthExternal || m_dwAuthCookie==NULL)
		{
			oAuth->CreateAuthProvider(&oAuthProvider,&m_dwAuthCookie);
			if (!oAuthProvider)
				sErr = _T("Error creating auth provider.");

			oAuthProvider->InitEnd(m_bAuthNonInteractive?VARIANT_TRUE:VARIANT_FALSE, 
									m_bAuthInteractiveOnly?VARIANT_TRUE:VARIANT_FALSE);
			if (m_sOptConfigDir.GetLength())
				oAuthProvider->put_OptConfigDir(Utf82Ole(m_sOptConfigDir));
			if (m_sDefUserName.GetLength())
				oAuthProvider->put_DefUserName(Utf82Ole(m_sDefUserName));
			if (m_sDefPassword.GetLength())
				oAuthProvider->put_DefPassword(Utf82Ole(m_sDefPassword));
		}
		else
		{
			oAuth->RetrieveAuthProvider(m_dwAuthCookie,&oAuthProvider);
			if (!oAuthProvider)
				sErr = _T("Error using last auth provider.");
		}
		*ppCookie = m_dwAuthCookie;
	}
	else
	{
		oAuth->RetrieveAuthProvider(nCookie,&oAuthProvider);
		*ppCookie = m_dwAuthCookie = nCookie;
		if (!oAuthProvider)
			sErr = _T("Error retriving auth provider.");
	}

	if (oAuthProvider)
	{
		svn_auth_baton_t *ab;
		oAuthProvider->get_SvnAuthBaton((void**)&ab);
		m_pCtx->auth_baton = ab;
		if (dwOldCookie) oAuth->ReleaseAuthProvider(dwOldCookie);
		m_oAuthProvider = oAuthProvider;
	}
	else 
	{
		m_dwAuthCookie = dwOldCookie;
		*ppCookie = dwOldCookie;
		return AtlReportError(__uuidof(SVNClient), sErr);
	}

	TRACELOG_END("_SetupAuth")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(S_OK)	TRACELOG_NEWLINE

	return S_OK;
}


STDMETHODIMP CSVNClient::SetupAuth(LONG	nCookie, LONG * ppCookie	)
{
	TRACELOG_BEGIN("SetupAuth")

	TRACELOG_TITLE("nCookie: ")	TRACELOG_NUMBER(nCookie)	TRACELOG_NEWLINE
	TRACELOG_TITLE("ppCookie: ")	TRACELOG_NUMBER(*ppCookie)	TRACELOG_NEWLINE

	HRESULT hr = InitClient(); if (FAILED(hr)) return hr;
	hr = _SetupAuth(nCookie, ppCookie);
	if (!FAILED(hr))
		m_bAuthExternal = true;

	TRACELOG_END("SetupAuth")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::AddFiles(IStringVector* arStrings, VARIANT_BOOL bRecurse, VARIANT_BOOL bForce, VARIANT_BOOL bNoIgnore, ISVNWcNotifyVector** ppStatus)
{
	TRACELOG_BEGIN("AddFiles")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecurse: ")		TRACELOG_NUMBER(bRecurse)					TRACELOG_NEWLINE
	TRACELOG_TITLE("bForce: ")		TRACELOG_NUMBER(bForce)						TRACELOG_NEWLINE
	TRACELOG_TITLE("bNoIgnore: ")	TRACELOG_NUMBER(bNoIgnore)					TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	CComPtr<ISVNWcNotifyVector> oStatus;
	oStatus.CoCreateInstance(__uuidof(SVNWcNotifyVector));
	m_oClientNotifyBaton.oNotifyVector = oStatus;

	long lCount;
	arStrings->get_Count(&lCount);
	apr_pool_t *subpool = svn_pool_create(pFuncPool);
	for (long i = 1; i <= lCount; i++)
	{
		svn_pool_clear(subpool);
		CComBSTR bstrFile; 
		arStrings->get_Item(i, &bstrFile);
		CString sFile = Ole2Uri(bstrFile, subpool);
		SVN_CALL_SVNERROR(svn_client_add3 (sFile, bRecurse==VARIANT_TRUE, bForce==VARIANT_TRUE, bNoIgnore==VARIANT_TRUE, m_pCtx, subpool));
	}
	svn_pool_destroy(subpool);

	oStatus.CopyTo(ppStatus);
	TRACELOG_TITLE("ppStatus: ")	TRACELOG_ISVN_WC_NOTIFY_VECTOR(*ppStatus)	TRACELOG_NEWLINE
	END_SVN_CALL

	m_oClientNotifyBaton.oNotifyVector = NULL;

	TRACELOG_END("AddFiles")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::GetFiles(IStringVector* arStrings, BSTR sRev, VARIANT_BOOL bRecurse, VARIANT_BOOL bIgnoreExternals, ISVNWcNotifyVector** ppStatus)
{
	TRACELOG_BEGIN("GetFiles")

	TRACELOG_TITLE("arStrings: ")		TRACELOG_ISTRING_VECTOR(arStrings)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sRev: ")				TRACELOG_WCHAR(sRev)						TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecurse: ")			TRACELOG_NUMBER(bRecurse)					TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreExternals: ")	TRACELOG_NUMBER(bIgnoreExternals)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	svn_opt_revision_t revision;
	string_to_revision(revision,Ole2Utf8(sRev));

	CComPtr<ISVNWcNotifyVector> oStatus;
	oStatus.CoCreateInstance(__uuidof(SVNWcNotifyVector));
	m_oClientNotifyBaton.oNotifyVector = oStatus;

	long lCount;
	arStrings->get_Count(&lCount);

	apr_array_header_t *arFiles;
	apr_array_header_t *arRevNums;
	arFiles = apr_array_make (pFuncPool,lCount, sizeof (const char *));



	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile; 
		arStrings->get_Item(i, &bstrFile);
		CString sFile = Ole2Uri(bstrFile, pFuncPool);
		const char * target = apr_pstrdup (pFuncPool, sFile);
        (*((const char **) apr_array_push (arFiles))) = target;
	}

	SVN_CALL_SVNERROR(svn_client_update2(&arRevNums, arFiles, &revision,
                                    bRecurse==VARIANT_TRUE, bIgnoreExternals==VARIANT_FALSE, m_pCtx, pFuncPool));

	oStatus.CopyTo(ppStatus);
	TRACELOG_TITLE("ppStatus: ")	TRACELOG_ISVN_WC_NOTIFY_VECTOR(*ppStatus)	TRACELOG_NEWLINE
	END_SVN_CALL

	m_oClientNotifyBaton.oNotifyVector = NULL;

	TRACELOG_END("GetFiles")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::List(BSTR sPath, BSTR sPRevision, BSTR sRevision, VARIANT_BOOL bRecurse, IStringVector** arStrings) 
{
	TRACELOG_BEGIN("List")

	TRACELOG_TITLE("sPath: ")		TRACELOG_WCHAR(sPath)					TRACELOG_NEWLINE
	TRACELOG_TITLE("sPRevision: ")	TRACELOG_WCHAR(sPRevision)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sRevision: ")	TRACELOG_WCHAR(sRevision)				TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecurse: ")	TRACELOG_NUMBER(bRecurse)				TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	CComPtr<IStringVector> sv;
	sv.CoCreateInstance(__uuidof(StringVector));

	apr_hash_t *dirents;
	svn_opt_revision_t peg_revision;
	svn_opt_revision_t revision;
	string_to_revision(peg_revision,Ole2Utf8(sPRevision));
	string_to_revision(revision,Ole2Utf8(sRevision));

	CString truePath = Ole2Uri(sPath, pFuncPool);
	SVN_CALL_SVNERROR(svn_client_ls2 (&dirents, truePath, &peg_revision, &revision,
						bRecurse == VARIANT_TRUE, m_pCtx, pFuncPool));

	apr_array_header_t *array;

	array = svn_sort__hash (dirents, svn_sort_compare_items_as_paths, pFuncPool);
	const char *utf8_entryname;
	svn_sort__item_t *item;
	svn_dirent_t *dirent;
	for (int i = 0; i < array->nelts; ++i)
    {
		item = &APR_ARRAY_IDX (array, i, svn_sort__item_t);
		utf8_entryname = (const char*)(item->key);

		dirent = (svn_dirent_t *)apr_hash_get (dirents, utf8_entryname, item->klen);
		if (dirent->kind == svn_node_dir)
		{
			CString sBuff = utf8_entryname;
			sBuff += "/";
			sv->Add(Utf82Ole(sBuff));
		}
		else
			sv->Add(Utf82Ole(utf8_entryname));
	}

	sv->QueryInterface(arStrings);
	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(*arStrings)		TRACELOG_NEWLINE
	END_SVN_CALL

	TRACELOG_END("List")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

CString normalize_comment(BSTR _sComment)
{
	if (!_sComment) return NULL;
	if (!_sComment[0]) return CString("");
	size_t i = 0;
	while (_sComment[i + 1]) {
		if (_sComment[i] == '\r')
			if (_sComment[i + 1] == '\n')
				wcscpy(&(_sComment[i]), &(_sComment[i + 1]));
			else _sComment[i] = '\n';
		i++;
	}
	if (_sComment[i] == '\r')
		_sComment[i] = '\n';
	return Ole2Utf8(_sComment);
}

// the notification function, forward call as events to COM client
void client_statusnotify_func(void *baton, const char *_path, svn_wc_status2_t *status)
{
	// safely obtain pointer to client classes
	if (!baton) return;
	CComPtr<ISVNWcStatusInfo> oClient(*((ISVNWcStatusInfo**)baton));
	if (!oClient) 
		return;

	CString path = _path;
	path.Replace("/./", "/");
	CComPtr<IStringMap> oStringStatus;
	CComPtr<IStringVector> oFiles;
	CComPtr<ISVNWcStatusCollection>	oObjStatus;
	oClient->get_Textual(&oStringStatus);
	oClient->get_Files(&oFiles);
	oClient->get_Binary(&oObjStatus);

	oFiles->Add(Utf82Ole(path));
	if (oStringStatus)
	{
		char word[20];
		status_to_string((char *)word, status);
		oStringStatus->Add(Utf82Ole(path),Utf82Ole(word));
	}

	// place to retrieve statuses
	CComPtr<ISVNWcStatus2> oStatus;
	HRESULT hr = oStatus.CoCreateInstance(__uuidof(SVNWcStatus));
	if (FAILED(hr))
		DebugBreak();
	oStatus->put_Copied(status->copied);
	oStatus->put_Locked(status->locked);
	oStatus->put_PropStatus((SvnWcStatusKind)status->prop_status);
	oStatus->put_ReposPropStatus((SvnWcStatusKind)status->repos_prop_status);
	oStatus->put_ReposTextStatus((SvnWcStatusKind)status->repos_text_status);
	oStatus->put_Switched(status->switched);
	oStatus->put_TextStatus((SvnWcStatusKind)status->text_status);

	CY cyTemp;

	oStatus->put_URL(Utf82Ole(status->url));
	oStatus->put_LastCmtRev(status->ood_last_cmt_rev);
	cyTemp.int64 = status->ood_last_cmt_date;
	oStatus->put_LastCmtDate(cyTemp);
	oStatus->put_LastCmtKind((SvnNodeKind)status->ood_kind);
	oStatus->put_LastCmdAuthor(Utf82Ole(status->ood_last_cmt_author));

	if (status->entry)
	{
		CComPtr<ISVNWcEntry> oEntry;
		HRESULT hr = oEntry.CoCreateInstance(__uuidof(SVNWcEntry));

		oEntry->put_Absent(status->entry->absent);
		oEntry->put_Checksum(Utf82Ole(status->entry->checksum));
		oEntry->put_CmtAuthor(Utf82Ole(status->entry->cmt_author));
		cyTemp.int64=status->entry->cmt_date;
		oEntry->put_CmtDate( cyTemp );
		oEntry->put_CmtRev(status->entry->cmt_rev);
		oEntry->put_ConflictNew(Utf82Ole(status->entry->conflict_new));
		oEntry->put_ConflictOld(Utf82Ole(status->entry->conflict_old));
		oEntry->put_ConflictWrk(Utf82Ole(status->entry->conflict_wrk));
		oEntry->put_Copied(status->entry->copied);
		oEntry->put_CopyFromRev(status->entry->copyfrom_rev);
		oEntry->put_CopyFromUrl(Utf82Ole(status->entry->copyfrom_url));
		oEntry->put_Deleted(status->entry->deleted);
		oEntry->put_Incomplete(status->entry->incomplete);
		oEntry->put_Kind((SvnNodeKind)status->entry->kind);
		oEntry->put_LockComment(Utf82Ole(status->entry->lock_comment));
		cyTemp.int64=status->entry->lock_creation_date;
		oEntry->put_LockCreationDate(cyTemp);
		oEntry->put_LockOwner(Utf82Ole(status->entry->lock_owner));
		oEntry->put_LockToken(Utf82Ole(status->entry->lock_token));
		oEntry->put_Name(Utf82Ole(status->entry->name));
		oEntry->put_Prejfile(Utf82Ole(status->entry->prejfile));
		cyTemp.int64=status->entry->prop_time;
		oEntry->put_PropTime(cyTemp);
		oEntry->put_Repos(Utf82Ole(status->entry->repos));
		oEntry->put_Revision(status->entry->revision);
		oEntry->put_Schedule((SvnWcSchedule)status->entry->schedule);
		cyTemp.int64=status->entry->text_time;
		oEntry->put_TextTime(cyTemp);
		oEntry->put_Url(Utf82Ole(status->entry->url));
		oEntry->put_Uuid(Utf82Ole(status->entry->uuid));
		oStatus->put_Entry(oEntry);
	}

	if (status->repos_lock)
	{
		CComPtr<ISVNLock> oReposLock;
		HRESULT hr = oReposLock.CoCreateInstance(__uuidof(SVNLock));

		oReposLock->put_Comment(Utf82Ole(status->repos_lock->comment));
		cyTemp.int64=status->repos_lock->creation_date;
		oReposLock->put_CreationDate(cyTemp);
		cyTemp.int64=status->repos_lock->expiration_date;
		oReposLock->put_ExpirationDate(cyTemp);
		oReposLock->put_IsDavComment(status->repos_lock->is_dav_comment);
		oReposLock->put_Owner(Utf82Ole(status->repos_lock->owner));
		oReposLock->put_Path(Utf82Ole(status->repos_lock->path));
		oReposLock->put_Token(Utf82Ole(status->repos_lock->token));
		oStatus->put_ReposLock(oReposLock);
	}
	oObjStatus->Add(Utf82Ole(path),oStatus);
}

STDMETHODIMP CSVNClient::GetStatus(BSTR sPath, VARIANT_BOOL bRecurse, BSTR sRev, VARIANT_BOOL bGetAll, VARIANT_BOOL bUpdate, VARIANT_BOOL bNoIgnore, VARIANT_BOOL bIgnoreExternals, ISVNWcStatusInfo **ppStatus)
{
	TRACELOG_BEGIN("GetStatus")

	TRACELOG_TITLE("sPath: ")			TRACELOG_WCHAR(sPath)				TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecurse: ")			TRACELOG_NUMBER(bRecurse)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sRev: ")				TRACELOG_WCHAR(sRev)				TRACELOG_NEWLINE
	TRACELOG_TITLE("bGetAll: ")			TRACELOG_NUMBER(bGetAll)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bUpdate: ")			TRACELOG_NUMBER(bUpdate)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bNoIgnore: ")		TRACELOG_NUMBER(bNoIgnore)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreExternals: ")	TRACELOG_NUMBER(bIgnoreExternals)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	// prepare baton
	CComPtr<ISVNWcStatusInfo> oStatusInfo;
	oStatusInfo.CoCreateInstance(__uuidof(SVNWcStatusInfo));

	TRACELOG_TITLE("GetStatus1") TRACELOG_NEWLINE //????

	CComPtr<IStringVector>	oFiles;
	oFiles.CoCreateInstance(__uuidof(StringVector));
	oStatusInfo->put_Files(oFiles);
	if (TRUE) 
	{
		CComPtr<IStringMap>	oStringStatus;
		oStringStatus.CoCreateInstance(__uuidof(StringMap));
		oStatusInfo->put_Textual(oStringStatus);
	}
	if (TRUE) 
	{
		CComPtr<ISVNWcStatusCollection>	oObjStatus;
		oObjStatus.CoCreateInstance(__uuidof(SVNWcStatusCollection));
		oStatusInfo->put_Binary(oObjStatus);
	}

	TRACELOG_TITLE("GetStatus2") TRACELOG_NEWLINE //????

	svn_revnum_t t; svn_opt_revision_t tt;

	CString sTemp = Ole2Utf8(sRev);
	string_to_revision(tt, sTemp.GetBuffer(sTemp.GetLength()) );

	TRACELOG_TITLE("GetStatus3") TRACELOG_NEWLINE //????

	CString sTempPath = Ole2Uri(sPath, pFuncPool);
	if ( bRecurse == VARIANT_FALSE && bUpdate == VARIANT_TRUE )
		sTempPath += "/.";

	TRACELOG_TITLE("GetStatus4") TRACELOG_NEWLINE //????

	 SVN_CALL_SVNERROR(svn_client_status2(&t,sTempPath,&tt,&client_statusnotify_func,&oStatusInfo.p,
		 bRecurse==VARIANT_TRUE,bGetAll==VARIANT_TRUE,bUpdate==VARIANT_TRUE,bNoIgnore==VARIANT_TRUE,
		 bIgnoreExternals==VARIANT_TRUE,m_pCtx,pFuncPool));

	// well done, here it is our collection

	TRACELOG_TITLE("GetStatus5") TRACELOG_NEWLINE //????

	oStatusInfo.CopyTo(ppStatus);
	#ifdef TRACE_LOG
		IStringVector* log_value;
		if(ppStatus != NULL)
			if(*ppStatus != NULL){
				(*ppStatus)->get_Files(&log_value);
				TRACELOG_TITLE("*ppStatus->Files: ")	TRACELOG_ISTRING_VECTOR(log_value)	TRACELOG_NEWLINE
			}
	#endif	

	TRACELOG_TITLE("GetStatus6") TRACELOG_NEWLINE //????

	END_SVN_CALL

	TRACELOG_TITLE("GetStatus7") TRACELOG_NEWLINE //????

	TRACELOG_END("GetStatus")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	TRACELOG_TITLE("GetStatus8") TRACELOG_NEWLINE //????

	return hr;
}

STDMETHODIMP CSVNClient::LockFiles(IStringVector* arStrings, BSTR sComment, VARIANT_BOOL bStealLock)
{
	TRACELOG_BEGIN("LockFiles")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sComment: ")		TRACELOG_WCHAR(sComment)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bStealLock: ")	TRACELOG_NUMBER(bStealLock)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	long lCount; 	
	arStrings->get_Count(&lCount);

	apr_array_header_t *files;
	files = apr_array_make (pFuncPool,lCount, sizeof (const char *));

	CString sPath;
	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile;
 		arStrings->get_Item(i, &bstrFile);
		sPath = Ole2Uri(bstrFile, pFuncPool);
		const char * target = apr_pstrdup (pFuncPool, sPath);
        (*((const char **) apr_array_push (files))) = target;
	}

    SVN_CALL_SVNERROR(svn_client_lock (files, normalize_comment(sComment), bStealLock == VARIANT_TRUE, m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("LockFiles")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::UnLockFiles(IStringVector* arStrings, VARIANT_BOOL bStealLock)
{
	TRACELOG_BEGIN("UnLockFiles")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bStealLock: ")	TRACELOG_NUMBER(bStealLock)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	long lCount; 	
	arStrings->get_Count(&lCount);

	apr_array_header_t *files;
	files = apr_array_make (pFuncPool,lCount, sizeof (const char *));

	CString sPath;
	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile;
 		arStrings->get_Item(i, &bstrFile);
		sPath = Ole2Uri(bstrFile, pFuncPool);
		const char * target = apr_pstrdup (pFuncPool, sPath);
        (*((const char **) apr_array_push (files))) = target;
	}

    SVN_CALL_SVNERROR(svn_client_unlock (files, bStealLock == VARIANT_TRUE, m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("UnLockFiles")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Revert(IStringVector* arStrings, VARIANT_BOOL bRecursive)
{
	TRACELOG_BEGIN("Revert")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")	TRACELOG_NUMBER(bRecursive)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	long lCount; 	
	arStrings->get_Count(&lCount);

	apr_array_header_t *files;
	files = apr_array_make (pFuncPool,lCount, sizeof (const char *));

	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile; 
		arStrings->get_Item(i, &bstrFile);
		CString sFile = Ole2Uri(bstrFile, pFuncPool);
		const char * target = apr_pstrdup (pFuncPool, sFile);
        (*((const char **) apr_array_push (files))) = target;
	}

	SVN_CALL_SVNERROR(svn_client_revert (files, bRecursive==VARIANT_TRUE, m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("Revert")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Delete(IStringVector* arStrings, BSTR sComment, VARIANT_BOOL bForce)
{
	TRACELOG_BEGIN("Delete")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sComment: ")		TRACELOG_WCHAR(sComment)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bForce: ")		TRACELOG_NUMBER(bForce)				TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	long lCount; 	
	arStrings->get_Count(&lCount);

	apr_array_header_t *files;
	files = apr_array_make (pFuncPool,lCount, sizeof (const char *));

	CString sPath;
	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile;
 		arStrings->get_Item(i, &bstrFile);
		sPath = Ole2Uri(bstrFile, pFuncPool);
		const char * target = apr_pstrdup (pFuncPool, sPath);
        (*((const char **) apr_array_push (files))) = target;
	}

	CString sTempComment = normalize_comment(sComment);
	m_pCtx->log_msg_baton = (void *)(LPCSTR)sTempComment;

	svn_client_commit_info_t *commit_info = NULL;
	SVN_CALL_SVNERROR(svn_client_delete (&commit_info, files, bForce == VARIANT_TRUE, m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("Delete")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::MkDir(IStringVector* arStrings, BSTR sComment)
{
	TRACELOG_BEGIN("MkDir")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sComment: ")		TRACELOG_WCHAR(sComment)			TRACELOG_NEWLINE

	int i=0;
	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	long lCount; 	
	arStrings->get_Count(&lCount);

	apr_array_header_t *files;
	files = apr_array_make (pFuncPool,lCount, sizeof (const char *));

	CString sPath;
	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile;
 		arStrings->get_Item(i, &bstrFile);
		sPath = Ole2Uri(bstrFile, pFuncPool);
		const char * target = apr_pstrdup (pFuncPool, sPath);
        (*((const char **) apr_array_push (files))) = target;
	}

	CString sTempComment = normalize_comment(sComment);
	m_pCtx->log_msg_baton = (void *)(LPCSTR)sTempComment;

	svn_client_commit_info_t *commit_info = NULL;
	SVN_CALL_SVNERROR(svn_client_mkdir (&commit_info, files, m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("MkDir")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::CleanUp(IStringVector* arStrings)
{
	TRACELOG_BEGIN("CleanUp")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	long lCount; 	
	arStrings->get_Count(&lCount);

	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile; 
		arStrings->get_Item(i, &bstrFile);
		svn_pool_clear (pFuncPool);
		CString sFile = Ole2Uri(bstrFile, pFuncPool);
		SVN_CALL_SVNERROR(svn_client_cleanup (sFile, m_pCtx, pFuncPool));
	}

	END_SVN_CALL

	TRACELOG_END("CleanUp")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Move(IStringVector* arFilesSrc, IStringVector* arFilesDst, VARIANT_BOOL bForce, BSTR sComment)
{
	TRACELOG_BEGIN("Move")

	TRACELOG_TITLE("arFilesSrc: ")	TRACELOG_ISTRING_VECTOR(arFilesSrc)	TRACELOG_NEWLINE
	TRACELOG_TITLE("arFilesDst: ")	TRACELOG_ISTRING_VECTOR(arFilesDst)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bForce: ")		TRACELOG_NUMBER(bForce)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sComment: ")		TRACELOG_WCHAR(sComment)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arFilesSrc || !arFilesDst) SVN_CALL_ERROR(E_INVALIDARG, "Vector(s) is Absent");

	long lCountSrc, lCountDst; 	
	arFilesSrc->get_Count(&lCountSrc);
	arFilesDst->get_Count(&lCountDst);
	if (lCountSrc != lCountDst) SVN_CALL_ERROR(E_INVALIDARG, "Various lengths of vectors.");

	CString sTempComment = Ole2Utf8(sComment);
	m_pCtx->log_msg_baton = (void *)(LPCSTR)sTempComment;

	CString sFileSrc, sFileDst;
	for (long i = 1; i <= lCountSrc; i++)
	{
		CComBSTR bstrFileSrc, bstrFileDst; 
		arFilesSrc->get_Item(i, &bstrFileSrc);
		arFilesDst->get_Item(i, &bstrFileDst);

		sFileSrc = Ole2Uri(bstrFileSrc, pFuncPool);
		sFileDst = Ole2Uri(bstrFileDst, pFuncPool);

		svn_commit_info_t *commit_info = NULL;
		SVN_CALL_SVNERROR(svn_client_move3 (&commit_info, sFileSrc, sFileDst,
								bForce == VARIANT_TRUE, m_pCtx, pFuncPool));
	}

	END_SVN_CALL

	TRACELOG_END("Move")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Resolved(IStringVector* arStrings, VARIANT_BOOL bRecursive)
{
	TRACELOG_BEGIN("Resolved")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")	TRACELOG_NUMBER(bRecursive)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	long lCount; 	
	arStrings->get_Count(&lCount);
	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile; 
		arStrings->get_Item(i, &bstrFile);
		svn_pool_clear (pFuncPool);
		CString sFile = Ole2Uri(bstrFile, pFuncPool);
		SVN_CALL_SVNERROR(svn_client_resolved (sFile, bRecursive==VARIANT_TRUE, m_pCtx, pFuncPool));
	}

	END_SVN_CALL

	TRACELOG_END("Resolved")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Commit(IStringVector* arStrings, BSTR sComment, VARIANT_BOOL bRecursive, VARIANT_BOOL bKeepLock)
{
	TRACELOG_BEGIN("Commit")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sComment: ")		TRACELOG_WCHAR(sComment)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")	TRACELOG_NUMBER(bRecursive)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bKeepLock: ")	TRACELOG_NUMBER(bKeepLock)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	long lCount; 	
	arStrings->get_Count(&lCount);

	apr_array_header_t *files;
	files = apr_array_make (pFuncPool,lCount, sizeof (const char *));

	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile; 
		arStrings->get_Item(i, &bstrFile);
		CString sFile = Ole2Uri(bstrFile, pFuncPool);
		const char * target = apr_pstrdup (pFuncPool, sFile);
        (*((const char **) apr_array_push (files))) = target;
	}

	CString sTempComment = normalize_comment(sComment);
	m_pCtx->log_msg_baton = (void *)(LPCSTR)sTempComment;

	svn_commit_info_t *commit_info = NULL;
	svn_error_t *__err = svn_client_commit3 (&commit_info, files, bRecursive==VARIANT_TRUE, bKeepLock==VARIANT_TRUE, m_pCtx, pFuncPool);
	SVN_CALL_SVNERROR(__err);

	END_SVN_CALL

	TRACELOG_END("Commit")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Copy(IStringVector* arFilesSrc, IStringVector* arFilesDst, BSTR sComment, BSTR sStartRevision)
{
	TRACELOG_BEGIN("Copy")

	TRACELOG_TITLE("arFilesSrc: ")		TRACELOG_ISTRING_VECTOR(arFilesSrc)		TRACELOG_NEWLINE
	TRACELOG_TITLE("arFilesDst: ")		TRACELOG_ISTRING_VECTOR(arFilesDst)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sComment: ")			TRACELOG_WCHAR(sComment)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRevision: ")	TRACELOG_WCHAR(sStartRevision)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arFilesSrc || !arFilesDst) SVN_CALL_ERROR(E_INVALIDARG, "Vector of files cannot be empty");

	svn_opt_revision_t revision;
	CString sTemp = Ole2Utf8(sStartRevision);
	string_to_revision(revision, sTemp.GetBuffer(sTemp.GetLength()+1));

	long lCountSrc, lCountDst; 	
	arFilesSrc->get_Count(&lCountSrc);
	arFilesDst->get_Count(&lCountDst);
	if (lCountSrc != lCountDst) SVN_CALL_ERROR(E_INVALIDARG, "Various lengths of vectors.");

	CString sLogMsg = normalize_comment(sComment);
	m_pCtx->log_msg_baton = (void *)(LPCSTR)sLogMsg;

	CString sFileSrc, sFileDst;
	for (long i = 1; i <= lCountSrc; i++)
	{
		CComBSTR bstrFileSrc, bstrFileDst; 
		arFilesSrc->get_Item(i, &bstrFileSrc);
		arFilesDst->get_Item(i, &bstrFileDst);

		sFileSrc = Ole2Uri(bstrFileSrc, pFuncPool);
		sFileDst = Ole2Uri(bstrFileDst, pFuncPool);

		svn_client_commit_info_t *commit_info = NULL;
		SVN_CALL_SVNERROR(svn_client_copy (&commit_info, sFileSrc, &revision, sFileDst, m_pCtx, pFuncPool));
	}

	END_SVN_CALL

	TRACELOG_END("Copy")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::CheckOut(IStringVector* arStrings, BSTR sPegRevision, BSTR sStartRevision, 
								  VARIANT_BOOL bRecursive, VARIANT_BOOL bIgnoreExternals)
{
	TRACELOG_BEGIN("CheckOut")

	TRACELOG_TITLE("arStrings: ")		TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sPegRevision: ")		TRACELOG_WCHAR(sPegRevision)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRevision: ")	TRACELOG_WCHAR(sStartRevision)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")		TRACELOG_NUMBER(bRecursive)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreExternals: ")	TRACELOG_NUMBER(bIgnoreExternals)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	if (!arStrings)
		SVN_CALL_ERROR(E_INVALIDARG,"No any files specified");

	long lCount; 	
	arStrings->get_Count(&lCount);

	svn_opt_revision_t peg_revision, revision;
	string_to_revision(peg_revision, Ole2Utf8(sPegRevision));
	string_to_revision(revision, Ole2Utf8(sStartRevision));
    if (revision.kind == svn_opt_revision_unspecified)
		revision.kind = svn_opt_revision_head;

	CComBSTR bstrLocalDir; 
	arStrings->get_Item(lCount, &bstrLocalDir);
	CString local_dir = Ole2Uri(bstrLocalDir, pFuncPool);
	if (svn_path_is_url (local_dir))
		{
			if (lCount == 1)
				{
					local_dir = svn_path_basename (local_dir, pFuncPool);
					local_dir = svn_path_uri_decode (local_dir, pFuncPool);
				}
			else
				local_dir = "";
		}
	else
	{
		if (lCount == 1)
			SVN_CALL_ERROR(E_INVALIDARG,"Putting local path as the last path you must provide at least one URL first");
		// the last path is local dir, exclude it from processing
		lCount --;
	}

	for (long i = 1; i <= lCount; i++)
	{
		svn_pool_clear (pFuncPool);

		CComBSTR bstrFile; 
		CString target_dir;

		arStrings->get_Item(i, &bstrFile);
		CString repos_url = Ole2Uri(bstrFile, pFuncPool);
		if (!svn_path_is_url (repos_url) )
			SVN_CALL_ERROR(E_FAIL,PushokAtlEx::FormatString("Path \"%s\" is not an URL",CW2T(bstrFile)));

		repos_url = svn_path_canonicalize (svn_path_uri_autoescape (repos_url, pFuncPool), pFuncPool);

		if (lCount == 1)
			{
				target_dir = local_dir;
			}
		else
			{
				target_dir = svn_path_basename (repos_url, pFuncPool);
				target_dir = svn_path_uri_decode (target_dir, pFuncPool);
				target_dir = svn_path_join (local_dir, target_dir, pFuncPool);
			}

		svn_revnum_t revnum;
		SVN_CALL_SVNERROR(svn_client_checkout2 (&revnum, repos_url, target_dir,
								&peg_revision,
								&revision,
								bRecursive==VARIANT_TRUE,
								bIgnoreExternals==VARIANT_TRUE,
								m_pCtx, pFuncPool));
	}

	END_SVN_CALL

	TRACELOG_END("CheckOut")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::RaCheckUrlValid(BSTR sUrl, BSTR sRev)
{
	TRACELOG_BEGIN("RaCheckUrlValid")

	TRACELOG_TITLE("sUrl: ")		TRACELOG_WCHAR(sUrl)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sRev: ")		TRACELOG_WCHAR(sRev)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_ra_session_t *ra_session;
	svn_revnum_t rev;
	svn_node_kind_t url_kind;
	const char *url;
    svn_opt_revision_t revision;
	string_to_revision(revision, Ole2Utf8(sRev));

	CString trueUrl = Ole2Uri(sUrl, pFuncPool);

	TRACELOG_TITLE("Pre_Session_from_path") TRACELOG_NEWLINE //????

	SVN_CALL_SVNERROR(svn_client__ra_session_from_path (&ra_session, &rev,
										&url, trueUrl, NULL, 
										&revision, &revision, m_pCtx, pFuncPool));

	TRACELOG_TITLE("Post_Session_from_path") TRACELOG_NEWLINE //????

	/* Decide if the URL is a file or directory. */
	SVN_CALL_SVNERROR(svn_ra_check_path (ra_session, "", rev, &url_kind, pFuncPool));

	TRACELOG_TITLE("Post_check_path") TRACELOG_NEWLINE //????

	if (url_kind==svn_node_none)
		SVN_CALL_ERROR(E_FAIL,"Connection to server exists but URL is neiver folder neither file");

	TRACELOG_TITLE("Post_check_url_kind") TRACELOG_NEWLINE //????

	END_SVN_CALL

	TRACELOG_END("RaCheckUrlValid")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::WcRestoreFiles(IStringVector* arStrings, VARIANT_BOOL bUseCommitTimes)
{
	TRACELOG_BEGIN("WcRestoreFiles")

	TRACELOG_TITLE("arStrings: ")		TRACELOG_ISTRING_VECTOR(arStrings)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bUseCommitTimes: ")	TRACELOG_NUMBER(bUseCommitTimes)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR (E_INVALIDARG, "Vector of files cannot be empty");

	const char *target;
	svn_wc_adm_access_t *adm_access;
	svn_wc_adm_access_t *dir_access;

	long lCount; 	
	arStrings->get_Count(&lCount);
	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrPath; 
		arStrings->get_Item(i, &bstrPath);
		CString sPath = Ole2Uri(bstrPath, pFuncPool);

		svn_pool_clear (pFuncPool);
		/* Use PATH to get the update's anchor and targets and get a write lock */
		if (svn_wc_adm_open_anchor (&adm_access, &dir_access, &target, sPath,
									TRUE, FALSE, m_pCtx->cancel_func, m_pCtx->cancel_baton,
									pFuncPool))
			continue;
		SVN_CALL_SVNERROR(restore_file (sPath, adm_access, bUseCommitTimes==VARIANT_TRUE, pFuncPool));
	}

	END_SVN_CALL

	TRACELOG_END("WcRestoreFiles")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

static svn_error_t *client_log_receiver (void *baton, apr_hash_t *changed_paths, svn_revnum_t rev,
										const char *author, const char *date, const char *msg,
										apr_pool_t *pool)
{
	::CoInitialize(NULL);
    CComObject<CSVNWcLogVector>* lv = (CComObject<CSVNWcLogVector>*)baton;

	// create notify action
	CComPtr<ISVNWcLog> oLog;
	oLog.CoCreateInstance(__uuidof(SVNWcLog));

	oLog->put_Revision(rev);
	oLog->put_Author(Utf82Ole(author));
	oLog->put_Message(Utf82Ole(msg));

	if (date && date[0])
	{
		/* Convert date to a format for humans. */
		apr_time_t time_temp;
		svn_time_from_cstring (&time_temp, date, pool);
		date = svn_time_to_human_cstring(time_temp, pool);
	}
	else
		date = "(no date)";
	oLog->put_Date(Utf82Ole(date));

	if (changed_paths)
  	{
      apr_array_header_t *sorted_paths;
      /* Get an array of sorted hash keys. */
      sorted_paths = svn_sort__hash (changed_paths, svn_sort_compare_items_as_paths, pool);
      for (int i = 0; i < sorted_paths->nelts; i++)
        {
          svn_sort__item_t *item = &(APR_ARRAY_IDX (sorted_paths, i, svn_sort__item_t));
          const char *path = (const char *)item->key;
          svn_log_changed_path_t *log_item = (svn_log_changed_path_t *)apr_hash_get (changed_paths, item->key, item->klen);

          CString copy_data = "";
          if (log_item->copyfrom_path && SVN_IS_VALID_REVNUM (log_item->copyfrom_rev))
			oLog->AddChangedPath(Utf82Ole(path), Utf82Ole(CString(log_item->action)), Utf82Ole(CString(log_item->copyfrom_path)), log_item->copyfrom_rev );
		  else
  			oLog->AddChangedPath(Utf82Ole(path), Utf82Ole(CString(log_item->action)), CComBSTR(""), 0);
        }
    }

	if (lv)
		lv->Add(oLog);

	// done using COM
	::CoUninitialize();
	return SVN_NO_ERROR;
}

STDMETHODIMP CSVNClient::Log(IStringVector* arStrings, BSTR sStartRev, BSTR sEndRev, LONG lLimit, 
							 VARIANT_BOOL bVerbose, VARIANT_BOOL bStopOnCopy, ISVNWcLogVector** arLogs)
{
	TRACELOG_BEGIN("Log")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRev: ")	TRACELOG_WCHAR(sStartRev)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sEndRev: ")		TRACELOG_WCHAR(sEndRev)					TRACELOG_NEWLINE
	TRACELOG_TITLE("lLimit: ")		TRACELOG_NUMBER(lLimit)					TRACELOG_NEWLINE
	TRACELOG_TITLE("bVerbose: ")		TRACELOG_NUMBER(bVerbose)				TRACELOG_NEWLINE
	TRACELOG_TITLE("bStopOnCopy: ")	TRACELOG_NUMBER(bStopOnCopy)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR (E_INVALIDARG, "Vector of files cannot be empty");

	CComPtr<ISVNWcLogVector> lv;
	lv.CoCreateInstance(__uuidof(SVNWcLogVector));

	svn_opt_revision_t start_revision, end_revision;
	string_to_revision(start_revision,Ole2Utf8(sStartRev));
	string_to_revision(end_revision,  Ole2Utf8(sEndRev));

	long lCount; 	
	arStrings->get_Count(&lCount);

	apr_array_header_t *files;
	files = apr_array_make (pFuncPool,lCount, sizeof (const char *));

	CString sPath;
	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile;
 		arStrings->get_Item(i, &bstrFile);
		sPath = Ole2Uri(bstrFile, pFuncPool);

		const char * target = apr_pstrdup (pFuncPool, sPath);

		if (lCount>1)
			if (i==1)
			{
				if ( !svn_path_is_url(target) )
					SVN_CALL_ERROR (E_FAIL, "When specifying working copy paths, only one target may be given");
			}
			else
				if ( svn_path_is_url(target) )
					SVN_CALL_ERROR (E_FAIL, "Only relative paths can be specified after a URL");

        (*((const char **) apr_array_push (files))) = target;
	}

    SVN_CALL_SVNERROR(svn_client_log2 (files, &start_revision, &end_revision, lLimit,
										bVerbose==VARIANT_TRUE, bStopOnCopy==VARIANT_TRUE, 
										client_log_receiver, (void *)lv, m_pCtx, pFuncPool));

	lv->QueryInterface(arLogs);
	TRACELOG_TITLE("arLogs: ")	TRACELOG_ISVN_WC_LOG_VECTOR(*arLogs)	TRACELOG_NEWLINE
	END_SVN_CALL

	TRACELOG_END("Log")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

/* A callback of type svn_info_receiver_t. */
static svn_error_t *client_info_receiver (void *baton, const char *path,
											const svn_info_t *info, apr_pool_t *pool)
{
	::CoInitialize(NULL);
    CComObject<CSVNWcInfoVector>* lv = (CComObject<CSVNWcInfoVector>*)baton;

	CComPtr<ISVNWcInfo> oInfo;
	oInfo.CoCreateInstance(__uuidof(SVNWcInfo));

	oInfo->put_Path(Utf82Ole(path));
	oInfo->put_Url(Utf82Ole(info->URL));
	oInfo->put_ReposRootUrl(Utf82Ole(info->repos_root_URL));
	oInfo->put_ReposUUID(Utf82Ole(info->repos_UUID));
	oInfo->put_LastChangedAuthor(Utf82Ole(info->last_changed_author));
	oInfo->put_CopyFromUrl(Utf82Ole(info->copyfrom_url));
	oInfo->put_Checksum(Utf82Ole(info->checksum));
	oInfo->put_ConflictOld(Utf82Ole(info->conflict_old));
	oInfo->put_ConflictNew(Utf82Ole(info->conflict_new));
	oInfo->put_ConflictWrk(Utf82Ole(info->conflict_wrk));
	oInfo->put_PrejFile(Utf82Ole(info->prejfile));
	oInfo->put_Revision(info->rev);
	oInfo->put_LastChangedRev(info->last_changed_rev);
	oInfo->put_CopyFromRev(info->copyfrom_rev);
	oInfo->put_HasWcInfo(info->has_wc_info);

	CY cyTemp;
	cyTemp.int64=info->last_changed_date;
	oInfo->put_LastChangedDate(cyTemp);
	cyTemp.int64=info->text_time;
	oInfo->put_TextTime(cyTemp);
	cyTemp.int64=info->prop_time;
	oInfo->put_PropTime(cyTemp);

	oInfo->put_Kind((SvnNodeKind)info->kind);
	oInfo->put_Schedule((SvnWcSchedule)info->schedule);

	if (info->lock)
	{
		CComPtr<ISVNLock> oLock;
		HRESULT hr = oLock.CoCreateInstance(__uuidof(SVNLock));

		oLock->put_Comment(Utf82Ole(info->lock->comment));
		cyTemp.int64=info->lock->creation_date;
		oLock->put_CreationDate(cyTemp);
		cyTemp.int64=info->lock->expiration_date;
		oLock->put_ExpirationDate(cyTemp);
		oLock->put_IsDavComment(info->lock->is_dav_comment);
		oLock->put_Owner(Utf82Ole(info->lock->owner));
		oLock->put_Path(Utf82Ole(info->lock->path));
		oLock->put_Token(Utf82Ole(info->lock->token));
		oInfo->put_Lock(oLock);
	}

	if (lv)
		lv->Add(oInfo);

	// done using COM
	::CoUninitialize();
	return SVN_NO_ERROR;
}

STDMETHODIMP CSVNClient::Info(IStringVector* arStrings, BSTR sPegRev, BSTR sStartRev, VARIANT_BOOL bRecursiv, 
							  ISVNWcInfoVector** arInfo)
{
	TRACELOG_BEGIN("Info")

	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISTRING_VECTOR(arStrings)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sPegRev: ")		TRACELOG_WCHAR(sPegRev)					TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRev: ")	TRACELOG_WCHAR(sStartRev)				TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursiv: ")	TRACELOG_NUMBER(bRecursiv)				TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arStrings) SVN_CALL_ERROR (E_INVALIDARG, "Vector of files cannot be empty");


	CComPtr<ISVNWcInfoVector> lv;
	lv.CoCreateInstance(__uuidof(SVNWcInfoVector));

	svn_opt_revision_t peg_revision, start_revision;
	string_to_revision(peg_revision, Ole2Utf8(sPegRev));
	string_to_revision(start_revision, Ole2Utf8(sStartRev));

	long lCount; 	
	arStrings->get_Count(&lCount);

	apr_array_header_t *files;
	files = apr_array_make (pFuncPool,lCount, sizeof (const char *));

	CString sPath;
	for (long i = 1; i <= lCount; i++)
	{
		CComBSTR bstrFile;
 		arStrings->get_Item(i, &bstrFile);
		sPath = Ole2Uri(bstrFile, pFuncPool);

		SVN_CALL_SVNERROR(svn_client_info (sPath, &peg_revision, &start_revision, client_info_receiver, 
								(void *)lv, bRecursiv==VARIANT_TRUE, m_pCtx, pFuncPool));
	}

	lv->QueryInterface(arInfo);
	TRACELOG_TITLE("arInfo: ")	TRACELOG_ISVN_WC_INFOVECTOR(*arInfo)	TRACELOG_NEWLINE
	END_SVN_CALL

	TRACELOG_END("Info")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::PropGet(BSTR sFile, BSTR sPropName, BSTR sPegRev, BSTR sStartRev, VARIANT_BOOL bRecursiv, BSTR* sVal)
{
	TRACELOG_BEGIN("PropGet")

	TRACELOG_TITLE("sFile: ")		TRACELOG_WCHAR(sFile)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sPropName: ")	TRACELOG_WCHAR(sPropName)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sPegRev: ")		TRACELOG_WCHAR(sPegRev)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRev: ")	TRACELOG_WCHAR(sStartRev)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursiv: ")	TRACELOG_NUMBER(bRecursiv)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_opt_revision_t peg_revision, start_revision;
	string_to_revision(peg_revision, Ole2Utf8(sPegRev));
	string_to_revision(start_revision, Ole2Utf8(sStartRev));
    apr_hash_t *props;

	CString sPath;
	sPath = Ole2Uri(sFile, pFuncPool);

	SVN_CALL_SVNERROR(svn_client_propget2 (&props, Ole2Utf8(sPropName), sPath, &peg_revision, 
							&start_revision, bRecursiv==VARIANT_TRUE, m_pCtx, pFuncPool));

    apr_hash_index_t *hi = apr_hash_first (pFuncPool, props);
    const void *key;
    svn_string_t *propval;

	CComBSTR bsrtVal;
	if (hi)
	{
		apr_hash_this (hi, &key, NULL, (void**)&propval);
		bsrtVal = Utf82Ole(propval->data);
	}
	else
		bsrtVal = Utf82Ole("");

	bsrtVal.CopyTo(sVal);
	END_SVN_CALL

	TRACELOG_END("PropGet")
	TRACELOG_TITLE("sVal: ")	TRACELOG_WCHAR(*sVal)	TRACELOG_NEWLINE
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)		TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::PropSet(BSTR sFile, BSTR sPropName, BSTR sVal, VARIANT_BOOL bRecursiv, VARIANT_BOOL bForce)
{
	TRACELOG_BEGIN("PropSet")

	TRACELOG_TITLE("sFile: ")		TRACELOG_WCHAR(sFile)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sPropName: ")	TRACELOG_WCHAR(sPropName)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sVal: ")			TRACELOG_WCHAR(sVal)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursiv: ")	TRACELOG_NUMBER(bRecursiv)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bForce: ")		TRACELOG_NUMBER(bForce)		TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_string_t *propval = svn_string_create (Ole2Utf8(sVal), pFuncPool);

	CString sProprtyName = Ole2Utf8(sPropName);
	if (sProprtyName=="FromFile")
	{
		//???? get property from file (path in propval)
	}

	CString sPath;
	sPath = Ole2Uri(sFile, pFuncPool);

	SVN_CALL_SVNERROR(svn_client_propset2 (sProprtyName, propval, sPath,
                             bRecursiv==VARIANT_TRUE, bForce==VARIANT_TRUE, m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("PropSet")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::PropDel(BSTR sFile, BSTR sPropName, VARIANT_BOOL bRecursiv)
{
	TRACELOG_BEGIN("PropDel")

	TRACELOG_TITLE("sFile: ")		TRACELOG_WCHAR(sFile)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sPropName: ")	TRACELOG_WCHAR(sPropName)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursiv: ")	TRACELOG_NUMBER(bRecursiv)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	CString sPath;
	sPath = Ole2Uri(sFile, pFuncPool);

	SVN_CALL_SVNERROR(svn_client_propset2 (Ole2Utf8(sPropName), NULL, sPath,
                             bRecursiv==VARIANT_TRUE, false, m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("PropDel")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::PropList(BSTR sFile, BSTR sPegRev, BSTR sStartRev, VARIANT_BOOL bRecursiv, ISVNWcPropListVector** arStrings)
{
	TRACELOG_BEGIN("PropList")

	TRACELOG_TITLE("sFile: ")		TRACELOG_WCHAR(sFile)							TRACELOG_NEWLINE
	TRACELOG_TITLE("sPegRev: ")		TRACELOG_WCHAR(sPegRev)							TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRev: ")	TRACELOG_WCHAR(sStartRev)						TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursiv: ")	TRACELOG_NUMBER(bRecursiv)						TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	CComPtr<ISVNWcPropListVector> oList;
	oList.CoCreateInstance(__uuidof(SVNWcPropListVector));

	svn_opt_revision_t peg_revision, start_revision;
	string_to_revision(peg_revision, Ole2Utf8(sPegRev));
	string_to_revision(start_revision, Ole2Utf8(sStartRev));
    apr_array_header_t *props;

	CString truePath = Ole2Uri(sFile, pFuncPool);

	SVN_CALL_SVNERROR(svn_client_proplist2 (&props, truePath, &peg_revision, &start_revision,
									bRecursiv==VARIANT_TRUE, m_pCtx, pFuncPool));

    for (int j = 0; j < props->nelts; ++j)
    {
		CComPtr<ISVNWcPropList> oProp;
		oProp.CoCreateInstance(__uuidof(SVNWcPropList));

        svn_client_proplist_item_t *item 
        = ((svn_client_proplist_item_t **)props->elts)[j];

		oProp->put_Path(Utf82Ole(item->node_name->data));

		apr_hash_index_t *hi;
		for (hi = apr_hash_first (pFuncPool, item->prop_hash); hi; hi = apr_hash_next (hi))
		{
			const void *key;
			void *val;
			apr_hash_this (hi, &key, NULL, &val);

			oProp->AddProp(Utf82Ole((const char *)key), Utf82Ole(((svn_string_t *)val)->data));
		}

		oList->Add(oProp);
    }

	oList->QueryInterface(arStrings);
	TRACELOG_TITLE("arStrings: ")	TRACELOG_ISVN_WC_PROPLIST_VECTOR(*arStrings)	TRACELOG_NEWLINE
	END_SVN_CALL

	TRACELOG_END("PropList")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::RevPropGet(BSTR sUrl, BSTR sPropName, BSTR sRev, BSTR *sVal )
{
	TRACELOG_BEGIN("RevPropGet")

	TRACELOG_TITLE("sUrl: ")			TRACELOG_WCHAR(sUrl)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sPropName: ")	TRACELOG_WCHAR(sPropName)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sRev: ")			TRACELOG_WCHAR(sRev)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sVal: ")			TRACELOG_WCHAR(*sVal)		TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_opt_revision_t revision;
	string_to_revision(revision, Ole2Utf8(sRev));
	svn_revnum_t t_revision;
	svn_string_t *propval;

	CString trueUrl = Ole2Uri(sUrl, pFuncPool);
	SVN_CALL_SVNERROR(svn_client_revprop_get (Ole2Utf8(sPropName), &propval, trueUrl, &revision, 
							&t_revision, m_pCtx, pFuncPool));

	CComBSTR bsrtVal;
	bsrtVal = Utf82Ole(propval->data);
	bsrtVal.CopyTo(sVal);

	END_SVN_CALL

	TRACELOG_END("RevPropGet")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::RevPropSet(BSTR sUrl, BSTR sPropName, BSTR sVal, BSTR sRev,  VARIANT_BOOL bForce)
{
	TRACELOG_BEGIN("RevPropSet")

	TRACELOG_TITLE("sUrl: ")			TRACELOG_WCHAR(sUrl)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sPropName: ")	TRACELOG_WCHAR(sPropName)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sVal: ")			TRACELOG_WCHAR(sVal)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sRev: ")			TRACELOG_WCHAR(sRev)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bForce: ")		TRACELOG_NUMBER(bForce)		TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_string_t *propval = svn_string_create (Ole2Utf8(sVal), pFuncPool);
	svn_opt_revision_t revision;
	string_to_revision(revision, Ole2Utf8(sRev));
	svn_revnum_t t_revision;

	CString trueUrl = Ole2Uri(sUrl, pFuncPool);
	SVN_CALL_SVNERROR(svn_client_revprop_set (Ole2Utf8(sPropName), propval, trueUrl,
                             &revision, &t_revision, bForce==VARIANT_TRUE, m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("RevPropSet")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::RevPropDel(BSTR sUrl, BSTR sPropName, BSTR sRev)
{
	TRACELOG_BEGIN("RevPropDel")

	TRACELOG_TITLE("sUrl: ")			TRACELOG_WCHAR(sUrl)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sPropName: ")	TRACELOG_WCHAR(sPropName)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sRev: ")			TRACELOG_WCHAR(sRev)		TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_opt_revision_t revision;
	string_to_revision(revision, Ole2Utf8(sRev));
	svn_revnum_t t_revision;

	CString trueUrl = Ole2Uri(sUrl, pFuncPool);
	SVN_CALL_SVNERROR(svn_client_revprop_set (Ole2Utf8(sPropName), NULL, trueUrl,
                             &revision, &t_revision, false, m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("RevPropDel")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::RevPropList(BSTR sUrl, BSTR sRev, IStringVector **arListing)
{
	TRACELOG_BEGIN("RevPropList")

	TRACELOG_TITLE("sUrl: ")			TRACELOG_WCHAR(sUrl)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sRev: ")			TRACELOG_WCHAR(sRev)				TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	CComPtr<IStringVector> sv;
	sv.CoCreateInstance(__uuidof(StringVector));

	svn_opt_revision_t revision;
	string_to_revision(revision, Ole2Utf8(sRev));
    apr_hash_t *dirents;
	svn_revnum_t t_revision;

	CString trueUrl = Ole2Uri(sUrl, pFuncPool);
	SVN_CALL_SVNERROR(svn_client_revprop_list (&dirents, trueUrl, &revision, &t_revision
									, m_pCtx, pFuncPool));

	apr_array_header_t *array;

	array = svn_sort__hash (dirents, svn_sort_compare_items_as_paths, pFuncPool);
	const char *utf8_entryname;
	svn_sort__item_t *item;
	svn_dirent_t *dirent;
	for (int i = 0; i < array->nelts; ++i)
    {
		item = &APR_ARRAY_IDX (array, i, svn_sort__item_t);
		utf8_entryname = (const char*)(item->key);

		dirent = (svn_dirent_t *)apr_hash_get (dirents, utf8_entryname, item->klen);
		if (dirent->kind == svn_node_dir)
		{
			CString sBuff = utf8_entryname;
			sBuff += "/";
			sv->Add(Utf82Ole(sBuff));
		}
		else
			sv->Add(Utf82Ole(utf8_entryname));
	}

	sv->QueryInterface(arListing);
	TRACELOG_TITLE("arListing: ")	TRACELOG_ISTRING_VECTOR(*arListing)	TRACELOG_NEWLINE
	END_SVN_CALL

	TRACELOG_END("RevPropList")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Version(LONG *nComSvnVersion, LONG *nSvnVersion, BSTR* sVal)
{
	TRACELOG_BEGIN("Version")

	TRACELOG_TITLE("nComSvnVersion: ")	TRACELOG_NUMBER(*nComSvnVersion)	TRACELOG_NEWLINE
	TRACELOG_TITLE("nSvnVersion: ")		TRACELOG_NUMBER(*nSvnVersion)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sVal: ")				TRACELOG_WCHAR(*sVal)		TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	*nComSvnVersion = 010406;
	*nSvnVersion = 010406;

	CComBSTR bstrRez;
	CString sTemp = "COM SVN Bindings, version 1.5.1.2\n";
	sTemp += "Copyright (C) 2006-2008 PushOk Software.\nCOM bindings is open source software, see\n"
			 "http://www.pushok.com/soft_svncom.php\n\n";
	sTemp += "svn, version ";
	sTemp += SVN_VERSION;
	sTemp += "\n   compiled ";
	sTemp += __DATE__;
	sTemp += ", ";
	sTemp += __TIME__;
	sTemp += "\n\n";

	sTemp += "Copyright (C) 2000-2006 CollabNet.\nSubversion is open source software, see"
                        " http://subversion.tigris.org/\nThis product includes software developed by "
                        "CollabNet (http://www.Collab.Net/).\n\n";

	svn_stringbuf_t *version_footer = 
		svn_stringbuf_create ("The following repository access (RA) modules are available:\n\n", pFuncPool);

	SVN_CALL_SVNERROR(svn_ra_print_modules (version_footer, pFuncPool));

	sTemp += version_footer->data;
	bstrRez = sTemp;

	bstrRez.CopyTo(sVal);
	END_SVN_CALL

	TRACELOG_END("Version")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

static svn_error_t *
client_blame_receiver (void *baton,
                apr_int64_t line_no,
                svn_revnum_t revision,
                const char *author,
                const char *date,
                const char *line,
                apr_pool_t *pool)
{
	::CoInitialize(NULL);
    CComObject<CSVNWcAnnVector>* lv = (CComObject<CSVNWcAnnVector>*)baton;

	// create notify action
	CComPtr<ISVNWcAnn> oAnn;
	oAnn.CoCreateInstance(__uuidof(SVNWcAnn));

	oAnn->put_Rev(revision);
	oAnn->put_Author(Utf82Ole(author));
	oAnn->put_Date(Utf82Ole(date));
	oAnn->put_Line(Utf82Ole(line));
	CY cyLineNo;
	cyLineNo.int64 = line_no;
	oAnn->put_LineNo(cyLineNo);

	if (lv)
		lv->Add(oAnn);

	// done using COM
	::CoUninitialize();
	return SVN_NO_ERROR;
}
 

/* This implements the `svn_opt_subcommand_t' interface. */
STDMETHODIMP CSVNClient::Annotate(BSTR sFile, BSTR sPegRev, BSTR sStartRev, BSTR sEndRev, ISVNWcAnnVector** arAnns)
{
	TRACELOG_BEGIN("Annotate")

	TRACELOG_TITLE("sFile: ")		TRACELOG_WCHAR(sFile)					TRACELOG_NEWLINE
	TRACELOG_TITLE("sPegRev: ")		TRACELOG_WCHAR(sPegRev)					TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRev: ")	TRACELOG_WCHAR(sStartRev)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sEndRev: ")		TRACELOG_WCHAR(sEndRev)					TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_opt_revision_t start_revision, end_revision, peg_revision;
	string_to_revision(start_revision, Ole2Utf8(sStartRev));
	string_to_revision(end_revision, Ole2Utf8(sEndRev));
	string_to_revision(peg_revision, Ole2Utf8(sPegRev));

	CString sPath = Ole2Uri(sFile, pFuncPool);
	sPath.Replace('\\', '/');

	CComPtr<ISVNWcAnnVector> lv;
	lv.CoCreateInstance(__uuidof(SVNWcAnnVector));
//	CComBSTR oRez;
	
    SVN_CALL_SVNERROR(svn_client_blame2 (sPath, &peg_revision, &start_revision, &end_revision, 
							client_blame_receiver, (void *)lv, m_pCtx, pFuncPool));

	lv->QueryInterface(arAnns);
	TRACELOG_TITLE("arAnns: ")	TRACELOG_ISVN_WC_ANN_VECTOR(*arAnns)	TRACELOG_NEWLINE
	END_SVN_CALL

	TRACELOG_END("Annotate")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return 	hr;
}

svn_error_t *client_write_cat(void *baton, const char *data, apr_size_t *len)
{
	CComBSTR *pFile = (CComBSTR *)baton;
	*pFile += Ansi2Ole(data, (int)*len);
    return SVN_NO_ERROR;
}

struct svn_stream_t {
  void *baton;
  svn_read_fn_t read_fn;
  svn_write_fn_t write_fn;
  svn_close_fn_t close_fn;
};

STDMETHODIMP CSVNClient::Cat(BSTR sFile, BSTR sPegRev, BSTR sStartRev, BSTR* sVal)
{
	TRACELOG_BEGIN("Cat")

	TRACELOG_TITLE("sFile: ")		TRACELOG_WCHAR(sFile)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sPegRev: ")		TRACELOG_WCHAR(sPegRev)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRev: ")	TRACELOG_WCHAR(sStartRev)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sVal: ")			TRACELOG_WCHAR(*sVal)		TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_opt_revision_t peg_revision, start_revision;
	string_to_revision(peg_revision, Ole2Utf8(sPegRev));
	string_to_revision(start_revision, Ole2Utf8(sStartRev));

	CComBSTR oRez;
	svn_stream_t *stream = (svn_stream_t *)apr_palloc (pFuncPool, sizeof (*stream));
	stream->baton = (void *)&oRez;
	stream->write_fn = client_write_cat;
	stream->read_fn = NULL;
	stream->close_fn = NULL;

	CString sPath = Ole2Uri(sFile, pFuncPool);
	sPath.Replace('\\', '/');

    SVN_CALL_SVNERROR(svn_client_cat2 (stream, sPath, &peg_revision, &start_revision, m_pCtx, pFuncPool));

	oRez.CopyTo(sVal);
	END_SVN_CALL

	TRACELOG_END("Cat")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return 	hr;
}

//Not tested
STDMETHODIMP CSVNClient::ExportFiles(IStringVector* arFilesSrc, IStringVector* arFilesDst, BSTR sPRevision, BSTR sRevision, VARIANT_BOOL bRecurse, VARIANT_BOOL bOverwrite, VARIANT_BOOL bIgnoreExternals)
{
	TRACELOG_BEGIN("ExportFiles")

	TRACELOG_TITLE("arFilesSrc: ")		TRACELOG_ISTRING_VECTOR(arFilesSrc)	TRACELOG_NEWLINE
	TRACELOG_TITLE("arFilesDst: ")		TRACELOG_ISTRING_VECTOR(arFilesDst)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sPRevision: ")		TRACELOG_WCHAR(sPRevision)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sRevision: ")		TRACELOG_WCHAR(sRevision)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecurse: ")			TRACELOG_NUMBER(bRecurse)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bOverwrite: ")		TRACELOG_NUMBER(bOverwrite)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreExternals: ")	TRACELOG_NUMBER(bIgnoreExternals)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arFilesSrc || !arFilesDst) SVN_CALL_ERROR(E_INVALIDARG, "Vector(s) is Absent");

	long lCountSrc, lCountDst;
	arFilesSrc->get_Count(&lCountSrc);
	arFilesDst->get_Count(&lCountDst);
	if (lCountSrc != lCountDst) SVN_CALL_ERROR(E_INVALIDARG, "Various lengths of vectors.");

	CString sFileSrc;
	for (long i = 1; i <= lCountSrc; i++)
	{
		CComBSTR bstrFileSrc, bstrFileDst; 
		arFilesSrc->get_Item(i, &bstrFileSrc);
		sFileSrc = Ole2Utf8(bstrFileSrc);
		arFilesDst->get_Item(i, &bstrFileDst);

		svn_opt_revision_t peg_revision;
		svn_opt_revision_t revision;
		string_to_revision(peg_revision,Ole2Utf8(sPRevision));
		string_to_revision(revision,Ole2Utf8(sRevision));

		SVN_CALL_SVNERROR(svn_client_export3 (NULL,
                    sFileSrc,
                    Ole2Uri(bstrFileDst, pFuncPool),
                    &peg_revision,
                    &revision,
                    bOverwrite == VARIANT_TRUE, 
                    bIgnoreExternals == VARIANT_TRUE,
                    bRecurse == VARIANT_TRUE,
                    NULL,
                    m_pCtx,
                    pFuncPool));
	}

	END_SVN_CALL

	TRACELOG_END("ExportFiles")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::ImportFiles(IStringVector *arFilesSrc, IStringVector *arFilesDst, BSTR sComment, VARIANT_BOOL bNonRecursive, VARIANT_BOOL bNoIgnore) 
{
	TRACELOG_BEGIN("ImportFiles")

	TRACELOG_TITLE("arFilesSrc: ")		TRACELOG_ISTRING_VECTOR(arFilesSrc)	TRACELOG_NEWLINE
	TRACELOG_TITLE("arFilesDst: ")		TRACELOG_ISTRING_VECTOR(arFilesDst)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sComment: ")			TRACELOG_WCHAR(sComment)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bNonRecursive: ")	TRACELOG_NUMBER(bNonRecursive)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bNoIgnore: ")		TRACELOG_NUMBER(bNoIgnore)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!arFilesSrc || !arFilesDst) SVN_CALL_ERROR(E_INVALIDARG, "Vector(s) is absent");

	long lCountSrc, lCountDst;
	arFilesSrc->get_Count(&lCountSrc);
	arFilesDst->get_Count(&lCountDst);
	if (lCountSrc != lCountDst) SVN_CALL_ERROR(E_INVALIDARG, "Various lengths of vectors.");

	if (!sComment) SVN_CALL_ERROR(E_INVALIDARG, "Coment is absent.");

	CString sTempComment = normalize_comment(sComment);
	m_pCtx->log_msg_baton = (void *)(LPCSTR)sTempComment;

	CString sFileDst;
	for (long i = 1; i <= lCountSrc; i++)
	{
		CComBSTR bstrFileSrc, bstrFileDst; 
		arFilesSrc->get_Item(i, &bstrFileSrc);
		arFilesDst->get_Item(i, &bstrFileDst);
		sFileDst = Ole2Uri(bstrFileDst, pFuncPool);

		svn_commit_info_t *commit_info = NULL;
		SVN_CALL_SVNERROR(svn_client_import2 (&commit_info,
                    Ole2Uri(bstrFileSrc, pFuncPool),
                    sFileDst,
                    bNonRecursive == VARIANT_TRUE,
                    bNoIgnore == VARIANT_TRUE,
                    m_pCtx,
                    pFuncPool));
	}

	END_SVN_CALL

	TRACELOG_END("ImportFiles")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Relocate(BSTR sPath, BSTR sFrom, BSTR sTo, VARIANT_BOOL bRecursive)
{
	TRACELOG_BEGIN("Relocate")

	TRACELOG_TITLE("sPath: ")		TRACELOG_WCHAR(sPath)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sFrom: ")		TRACELOG_WCHAR(sFrom)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sTo: ")			TRACELOG_WCHAR(sTo)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")	TRACELOG_NUMBER(bRecursive)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!sPath || !sFrom || sTo) SVN_CALL_ERROR(E_INVALIDARG, "Path is absent.");

	CString 
		sFromURL = Ole2Uri(sFrom, pFuncPool), 
		sToURL = Ole2Uri(sTo, pFuncPool);

	SVN_CALL_SVNERROR(svn_client_relocate (Ole2Uri(sPath, pFuncPool),
				sFromURL,
                sToURL,
                bRecursive == VARIANT_TRUE,
                m_pCtx,
                pFuncPool));

	END_SVN_CALL

	TRACELOG_END("Relocate")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Diff(BSTR sOptions, BSTR sPath1, BSTR sRevision1, BSTR sPath2, BSTR sRevision2,
				BSTR sOutFile, BSTR sErrFile, BSTR sHeaderEncoding, 
				VARIANT_BOOL bRecursive, VARIANT_BOOL bIgnoreAncestry, VARIANT_BOOL bNoDiffDeleted, VARIANT_BOOL bIgnoreContentType)
{
	TRACELOG_BEGIN("Diff")

	TRACELOG_TITLE("sOptions: ")				TRACELOG_WCHAR(sOptions)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sPath1: ")				TRACELOG_WCHAR(sPath1)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sRevision1: ")			TRACELOG_WCHAR(sRevision1)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sPath2: ")				TRACELOG_WCHAR(sPath2)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sRevision2: ")			TRACELOG_WCHAR(sRevision2)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sOutFile: ")				TRACELOG_WCHAR(sOutFile)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sErrFile: ")				TRACELOG_WCHAR(sErrFile)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sHeaderEncoding: ")		TRACELOG_WCHAR(sHeaderEncoding)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")			TRACELOG_NUMBER(bRecursive)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreAncestry: ")		TRACELOG_NUMBER(bIgnoreAncestry)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bNoDiffDeleted: ")		TRACELOG_NUMBER(bNoDiffDeleted)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreContentType: ")	TRACELOG_NUMBER(bIgnoreContentType)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!sOutFile) SVN_CALL_ERROR(E_INVALIDARG, "Out file is absent.");

	apr_file_t *outfile = NULL;
	apr_file_t *errfile = NULL;
	apr_status_t apr_err;
	if (sOutFile) {
		apr_err = apr_file_open (&outfile, Ole2Uri(sOutFile, pFuncPool),
                  (APR_WRITE | APR_TRUNCATE | APR_CREATE),
                  APR_OS_DEFAULT, pFuncPool);
		if (apr_err)
			SVN_CALL_ERROR(E_INVALIDARG, "Error open output file.");
	}
	if (sErrFile) {
		apr_err = apr_file_open (&errfile, Ole2Uri(sErrFile, pFuncPool),
                  (APR_WRITE | APR_TRUNCATE | APR_CREATE),
                  APR_OS_DEFAULT, pFuncPool);
		if (apr_err)
			SVN_CALL_ERROR(E_INVALIDARG, "Error open errfile file.");
	}

	svn_opt_revision_t revision1;
	string_to_revision(revision1,Ole2Utf8(sRevision1));
	svn_opt_revision_t revision2;
	string_to_revision(revision2,Ole2Utf8(sRevision2));

	apr_array_header_t *options = NULL;
    options = svn_cstring_split (Ole2Utf8(sOptions), " \t\n\r", TRUE, pFuncPool);

	CString 
		sPath1URL = Ole2Uri(sPath1, pFuncPool), 
		sPath2URL = Ole2Uri(sPath2, pFuncPool);


	SVN_CALL_SVNERROR(svn_client_diff3 (options,
				sPath1URL,
                &revision1,
                sPath2URL,
                &revision2,
                bRecursive == VARIANT_TRUE,
                bIgnoreAncestry == VARIANT_TRUE,
                bNoDiffDeleted == VARIANT_TRUE,
                bIgnoreContentType == VARIANT_TRUE,
				sHeaderEncoding?apr_pstrdup(pFuncPool, Ole2Utf8(sHeaderEncoding)):APR_LOCALE_CHARSET,
                outfile,
                errfile,
                m_pCtx,
                pFuncPool));
	END_SVN_CALL

	TRACELOG_END("Diff")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::DiffPeg(BSTR sOptions, BSTR sPath, BSTR sPegRevision, BSTR sStartRevision, BSTR sEndRevision,
				BSTR sOutFile, BSTR sErrFile, BSTR sHeaderEncoding, 
				VARIANT_BOOL bRecursive, VARIANT_BOOL bIgnoreAncestry, VARIANT_BOOL bNoDiffDeleted, VARIANT_BOOL bIgnoreContentType)
{
	TRACELOG_BEGIN("DiffPeg")

	TRACELOG_TITLE("sOptions: ")				TRACELOG_WCHAR(sOptions)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sPath: ")				TRACELOG_WCHAR(sPath)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sPegRevision: ")			TRACELOG_WCHAR(sPegRevision)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRevision: ")		TRACELOG_WCHAR(sStartRevision)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sEndRevision: ")			TRACELOG_WCHAR(sEndRevision)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sOutFile: ")				TRACELOG_WCHAR(sOutFile)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sErrFile: ")				TRACELOG_WCHAR(sErrFile)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sHeaderEncoding: ")		TRACELOG_WCHAR(sHeaderEncoding)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")			TRACELOG_NUMBER(bRecursive)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreAncestry: ")		TRACELOG_NUMBER(bIgnoreAncestry)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bNoDiffDeleted: ")		TRACELOG_NUMBER(bNoDiffDeleted)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreContentType: ")	TRACELOG_NUMBER(bIgnoreContentType)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL
	if (!sOutFile) SVN_CALL_ERROR(E_INVALIDARG, "Out file is absent.");

	apr_file_t *outfile = NULL;
	apr_file_t *errfile = NULL;
	apr_status_t apr_err;
	if (sOutFile) {
		apr_err = apr_file_open (&outfile, Ole2Uri(sOutFile, pFuncPool),
                  (APR_WRITE | APR_TRUNCATE | APR_CREATE),
                  APR_OS_DEFAULT, pFuncPool);
		if (apr_err)
			SVN_CALL_ERROR(E_INVALIDARG, "Error open output file.");
	}
	if (sErrFile) {
		apr_err = apr_file_open (&errfile, Ole2Uri(sErrFile, pFuncPool),
                  (APR_WRITE | APR_TRUNCATE | APR_CREATE),
                  APR_OS_DEFAULT, pFuncPool);
		if (apr_err)
			SVN_CALL_ERROR(E_INVALIDARG, "Error open errfile file.");
	}

	svn_opt_revision_t peg_revision;
	string_to_revision(peg_revision,Ole2Utf8(sPegRevision));
	svn_opt_revision_t start_revision;
	string_to_revision(start_revision,Ole2Utf8(sStartRevision));
	svn_opt_revision_t end_revision;
	string_to_revision(end_revision,Ole2Utf8(sEndRevision));

	apr_array_header_t *options = NULL;
    options = svn_cstring_split (Ole2Utf8(sOptions), " \t\n\r", TRUE, pFuncPool);

	CString 
		sPathURL = Ole2Uri(sPath, pFuncPool);

	SVN_CALL_SVNERROR(svn_client_diff_peg3 (options,
				sPathURL,
                &peg_revision,
                &start_revision,
                &end_revision,
                bRecursive == VARIANT_TRUE,
                bIgnoreAncestry == VARIANT_TRUE,
                bNoDiffDeleted == VARIANT_TRUE,
                bIgnoreContentType == VARIANT_TRUE,
				sHeaderEncoding?apr_pstrdup(pFuncPool, Ole2Utf8(sHeaderEncoding)):APR_LOCALE_CHARSET,
                outfile,
                errfile,
                m_pCtx,
                pFuncPool));
	END_SVN_CALL

	TRACELOG_END("DiffPeg")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::Merge(BSTR sPath1, BSTR sRevision1, BSTR sPath2, BSTR sRevision2, BSTR sTargetWCPath,
				VARIANT_BOOL bRecursive, VARIANT_BOOL bIgnoreAncestry, VARIANT_BOOL bForce, VARIANT_BOOL bDryRun) 
{
	TRACELOG_BEGIN("Merge")

	TRACELOG_TITLE("sPath1: ")			TRACELOG_WCHAR(sPath1)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sRevision1: ")		TRACELOG_WCHAR(sRevision1)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sPath2: ")			TRACELOG_WCHAR(sPath2)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sRevision2: ")		TRACELOG_WCHAR(sRevision2)			TRACELOG_NEWLINE
	TRACELOG_TITLE("sTargetWCPath: ")	TRACELOG_WCHAR(sTargetWCPath)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")		TRACELOG_NUMBER(bRecursive)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreAncestry: ")	TRACELOG_NUMBER(bIgnoreAncestry)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bForce: ")			TRACELOG_NUMBER(bForce)				TRACELOG_NEWLINE
	TRACELOG_TITLE("bDryRun: ")			TRACELOG_NUMBER(bDryRun)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_opt_revision_t revision1;
	string_to_revision(revision1,Ole2Utf8(sRevision1));
	svn_opt_revision_t revision2;
	string_to_revision(revision2,Ole2Utf8(sRevision2));

	CString 
		sPath1URL = Ole2Uri(sPath1, pFuncPool), 
		sPath2URL = Ole2Uri(sPath2, pFuncPool);

	SVN_CALL_SVNERROR(svn_client_merge (sPath1URL,
                &revision1,
                sPath2URL,
                &revision2,
				Ole2Uri(sTargetWCPath, pFuncPool),
                bRecursive == VARIANT_TRUE,
                bIgnoreAncestry == VARIANT_TRUE,
                bForce == VARIANT_TRUE,
                bDryRun == VARIANT_TRUE,
                m_pCtx,
                pFuncPool));

	END_SVN_CALL

	TRACELOG_END("Merge")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::MergePeg(BSTR sSource, BSTR sPegRevision, BSTR sStartRevision, BSTR sEndRevision, BSTR sTargetWCPath,
				VARIANT_BOOL bRecursive, VARIANT_BOOL bIgnoreAncestry, VARIANT_BOOL bForce, VARIANT_BOOL bDryRun)
{
	TRACELOG_BEGIN("MergePeg")

	TRACELOG_TITLE("sSource: ")			TRACELOG_WCHAR(sSource)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sPegRevision: ")		TRACELOG_WCHAR(sPegRevision)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sStartRevision: ")	TRACELOG_WCHAR(sStartRevision)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sEndRevision: ")		TRACELOG_WCHAR(sEndRevision)		TRACELOG_NEWLINE
	TRACELOG_TITLE("sTargetWCPath: ")	TRACELOG_WCHAR(sTargetWCPath)		TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")		TRACELOG_NUMBER(bRecursive)			TRACELOG_NEWLINE
	TRACELOG_TITLE("bIgnoreAncestry: ")	TRACELOG_NUMBER(bIgnoreAncestry)	TRACELOG_NEWLINE
	TRACELOG_TITLE("bForce: ")			TRACELOG_NUMBER(bForce)				TRACELOG_NEWLINE
	TRACELOG_TITLE("bDryRun: ")			TRACELOG_NUMBER(bDryRun)			TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_opt_revision_t peg_revision;
	string_to_revision(peg_revision,Ole2Utf8(sPegRevision));
	svn_opt_revision_t start_revision;
	string_to_revision(start_revision,Ole2Utf8(sStartRevision));
	svn_opt_revision_t end_revision;
	string_to_revision(end_revision,Ole2Utf8(sEndRevision));

	CString 
		sPathURL = Ole2Uri(sSource, pFuncPool);

	SVN_CALL_SVNERROR(svn_client_merge_peg (sPathURL,
                &peg_revision,
                &start_revision,
                &end_revision,
				Ole2Uri(sTargetWCPath, pFuncPool),
                bRecursive == VARIANT_TRUE,
                bIgnoreAncestry == VARIANT_TRUE,
                bForce == VARIANT_TRUE,
                bDryRun == VARIANT_TRUE,
                m_pCtx,
                pFuncPool));

	END_SVN_CALL

	TRACELOG_END("MergePeg")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::SetOption(SvnClientOption eOption, VARIANT vVal)
{
	TRACELOG_BEGIN("SetOption")

	TRACELOG_TITLE("eOption: ")		TRACELOG_NUMBER(eOption)		TRACELOG_NEWLINE
	TRACELOG_TITLE("vVal.vt: ")		TRACELOG_NUMBER(vVal.vt)		TRACELOG_NEWLINE
	TRACELOG_TITLE("vVal.bstrVal: ")	TRACELOG_WCHAR(vVal.bstrVal)	TRACELOG_NEWLINE
	TRACELOG_TITLE("vVal.boolVal: ")	TRACELOG_NUMBER(vVal.boolVal)	TRACELOG_NEWLINE

	HRESULT hr=S_OK;
	{
		Lock();
		m_bCancelRequested = false;
		try
		{
			switch (eOption){
			case SvnClientOptionAdmDir:
				if (vVal.vt == VT_BSTR)
					SVN_CALL_SVNERROR(svn_wc_set_adm_dir( Ole2Utf8(vVal.bstrVal),m_pPool))
				else
					SVN_CALL_ERROR(E_INVALIDARG, "Option \"AdmDir\" requires string value");
				break;

			case SvnClientOptionAppartReentrantEvents:
				if (vVal.vt == VT_BOOL)
					m_bGitEvents  = ! (vVal.boolVal == VARIANT_TRUE);
				else
					SVN_CALL_ERROR(E_INVALIDARG, "Option \"AppartReentrantEvents\" requires boolean value");
				break;

			case SvnClientOptionDefUserName:
				if (vVal.vt == VT_BSTR)
					m_sDefUserName = Ole2Utf8(vVal.bstrVal);
				else
					SVN_CALL_ERROR(E_INVALIDARG, "Option \"DefUserName\" requires string value");
				break;

			case SvnClientOptionDefPassword:
				if (vVal.vt == VT_BSTR)
					m_sDefPassword = Ole2Utf8(vVal.bstrVal);
				else
					SVN_CALL_ERROR(E_INVALIDARG, "Option \"DefPassword\" requires string value");
				break;

			case SvnClientOptionConfigDir:
				if (vVal.vt == VT_BSTR)
					m_sOptConfigDir = Ole2Utf8(vVal.bstrVal);
				else
					SVN_CALL_ERROR(E_INVALIDARG, "Option \"ConfigDir\" requires string value");
				break;

			case SvnClientOptionAuthInteractiveOnly:
				if (vVal.vt == VT_BOOL)
					m_bAuthInteractiveOnly  = vVal.boolVal == VARIANT_TRUE;
				else
					SVN_CALL_ERROR(E_INVALIDARG, "Option \"AuthNonInteractive\" requires boolean value");
				break;

			case SvnClientOptionAuthNonInteractive:
				if (vVal.vt == VT_BOOL)
					m_bAuthNonInteractive  = vVal.boolVal == VARIANT_TRUE;
				else
					SVN_CALL_ERROR(E_INVALIDARG, "Option \"AuthNonInteractive\" requires boolean value");
				break;

			default:
				SVN_CALL_ERROR(E_INVALIDARG, "Unknown option");
			}

		}catch (...){
			ATLASSERT(FALSE);
			hr = E_UNEXPECTED;
		}
		end_svn_call:
		Unlock();
	}

	TRACELOG_END("SetOption")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::GetOption(SvnClientOption eOption, VARIANT *pVal)
{
	TRACELOG_BEGIN("GetOption")

	TRACELOG_TITLE("eOption: ")			TRACELOG_NUMBER(eOption)		TRACELOG_NEWLINE
	TRACELOG_TITLE("pVal->vt: ")			TRACELOG_NUMBER(pVal->vt)		TRACELOG_NEWLINE
	TRACELOG_TITLE("pVal->bstrVal: ")	TRACELOG_WCHAR(pVal->bstrVal)	TRACELOG_NEWLINE
	TRACELOG_TITLE("pVal->boolVal: ")	TRACELOG_NUMBER(pVal->boolVal)	TRACELOG_NEWLINE

	HRESULT hr=S_OK;
	{
		Lock();
		m_bCancelRequested = false;
		try
		{
			if (pVal==NULL) SVN_CALL_ERROR(E_INVALIDARG, "Can not found variable for result return.");

			switch (eOption){
			case SvnClientOptionAdmDir:
				Utf82Ole(svn_wc_get_adm_dir(m_pPool)).CopyTo(pVal);
				break;

			case SvnClientOptionAppartReentrantEvents:
				pVal->boolVal = m_bGitEvents?VARIANT_FALSE:VARIANT_TRUE;
				break;

			case SvnClientOptionDefUserName:
				Utf82Ole(m_sDefUserName).CopyTo(pVal);
				break;

			case SvnClientOptionDefPassword:
				Utf82Ole(m_sDefPassword).CopyTo(pVal);
				break;

			case SvnClientOptionConfigDir:
				Utf82Ole(m_sOptConfigDir).CopyTo(pVal);
				break;

			case SvnClientOptionAuthInteractiveOnly:
				pVal->boolVal = m_bAuthInteractiveOnly?VARIANT_TRUE:VARIANT_FALSE;
				break;

			case SvnClientOptionAuthNonInteractive:
				pVal->boolVal = m_bAuthNonInteractive?VARIANT_TRUE:VARIANT_FALSE;
				break;

			default:
				SVN_CALL_ERROR(E_INVALIDARG, "Unknown option");
			}

		}catch (...){
			ATLASSERT(FALSE);
			hr = E_UNEXPECTED;
		}
		end_svn_call:
		Unlock();
	}

	TRACELOG_END("GetOption")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

STDMETHODIMP CSVNClient::RaCheckUrlKind(BSTR sUrl, BSTR sRev, SvnNodeKind *eKind)
{
	TRACELOG_BEGIN("RaCheckUrlKind")

	TRACELOG_TITLE("sUrl: ")		TRACELOG_WCHAR(sUrl)	TRACELOG_NEWLINE
	TRACELOG_TITLE("sRev: ")		TRACELOG_WCHAR(sRev)	TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	if (!eKind) SVN_CALL_ERROR(E_INVALIDARG, "Return value cannot be empty");

	svn_ra_session_t *ra_session;
	svn_revnum_t rev;
	svn_node_kind_t url_kind;
	const char *url;
    svn_opt_revision_t revision;
	string_to_revision(revision, Ole2Utf8(sRev));

	CString trueUrl = Ole2Uri(sUrl, pFuncPool);
	SVN_CALL_SVNERROR(svn_client__ra_session_from_path (&ra_session, &rev,
										&url, trueUrl,  NULL, 
										&revision, &revision, m_pCtx, pFuncPool));

	/* Decide if the URL is a file or directory. */
	SVN_CALL_SVNERROR(svn_ra_check_path (ra_session, "", rev, &url_kind, pFuncPool));
	if (url_kind==svn_node_none)
		SVN_CALL_ERROR(E_FAIL,"Connection to server exists but URL is neiver folder neither file");

	switch (url_kind) {
		case svn_node_none: *eKind = SvnNodeKindNone; break;
		case svn_node_file: *eKind = SvnNodeKindFile; break;
		case svn_node_dir: *eKind = SvnNodeKindDir; break;
		case svn_node_unknown: *eKind = SvnNodeKindUnknown; break;
		default: *eKind = SvnNodeKindUnknown;
	}
	END_SVN_CALL

	TRACELOG_END("RaCheckUrlKind")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}


STDMETHODIMP CSVNClient::RaGetReposiotryRoot(BSTR sPath, BSTR *sRoot) {
	TRACELOG_BEGIN("RaGetReposiotryRoot")

	TRACELOG_TITLE("sPath: ")			TRACELOG_WCHAR(sPath)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sRoot: ")			TRACELOG_WCHAR(*sRoot)				TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	svn_ra_session_t *ra_session;
	svn_revnum_t rev;
	const char *url;
	const char *repos_root;
    svn_opt_revision_t revision;
	string_to_revision(revision, Ole2Utf8(CComBSTR("HEAD")));

	CString trueUrl = Ole2Uri(sPath, pFuncPool);
	SVN_CALL_SVNERROR(svn_client__ra_session_from_path (&ra_session, &rev,
										&url, trueUrl,  NULL, 
										&revision, &revision, m_pCtx, pFuncPool));
	/* Getting repository root. */
	SVN_CALL_SVNERROR(svn_ra_get_repos_root (ra_session, &repos_root, pFuncPool));

	CComBSTR(repos_root).CopyTo(sRoot);
	END_SVN_CALL

	TRACELOG_END("RaGetReposiotryRoot")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}

struct svn_wc_adm_access_t
{
  /* PATH to directory which contains the administrative area */
  const char *path;

  enum svn_wc__adm_access_type {

    /* SVN_WC__ADM_ACCESS_UNLOCKED indicates no lock is held allowing
       read-only access */
    svn_wc__adm_access_unlocked,

    /* SVN_WC__ADM_ACCESS_WRITE_LOCK indicates that a write lock is held
       allowing read-write access */
    svn_wc__adm_access_write_lock,

    /* SVN_WC__ADM_ACCESS_CLOSED indicates that the baton has been
       closed. */
    svn_wc__adm_access_closed

  } type;

  /* LOCK_EXISTS is set TRUE when the write lock exists */
  svn_boolean_t lock_exists;

  /* SET_OWNER is TRUE if SET is allocated from this access baton */
  svn_boolean_t set_owner;

  /* The working copy format version number for the directory */
  int wc_format;

  /* SET is a hash of svn_wc_adm_access_t* keyed on char* representing the
     path to directories that are open. */
  apr_hash_t *set;

  /* ENTRIES is the cached entries for PATH, without those in state
     deleted. ENTRIES_HIDDEN is the cached entries including those in
     state deleted or state absent. Either may be NULL. */
  apr_hash_t *entries;
  apr_hash_t *entries_hidden;

  /* A hash mapping const char * entry names to hashes of wcprops.
     These hashes map const char * names to svn_string_t * values.
     NULL of the wcprops hasn't been read into memory.
     ### Since there are typically just one or two wcprops per entry,
     ### we could use a more compact way of storing them. */
  apr_hash_t *wcprops;

  /* POOL is used to allocate cached items, they need to persist for the
     lifetime of this access baton */
  apr_pool_t *pool;

};

STDMETHODIMP CSVNClient::LinkWithParentFolder(BSTR sPath)
{
	BEGIN_SVN_CALL
	
	CString stPath = Ole2Uri(sPath, pFuncPool);
	CString stFolder = stPath.Mid(stPath.ReverseFind('/')+1);
	CString stParent = stPath.Left(stPath.ReverseFind('/'));

	apr_hash_t *entries;
    svn_wc_adm_access_t *dir_access = (svn_wc_adm_access_t *)memset(apr_palloc(pFuncPool, sizeof(*dir_access)), 0, sizeof(*dir_access));
	dir_access->type = svn_wc_adm_access_t::svn_wc__adm_access_write_lock;
	dir_access->path = apr_pstrdup(pFuncPool, stParent);
	dir_access->pool = pFuncPool;
	dir_access->wc_format = SVN_WC__VERSION;

	SVN_CALL_SVNERROR(svn_wc_entries_read(&entries, dir_access, FALSE, pFuncPool));

    svn_wc_entry_t *new_entry = (svn_wc_entry_t *)memset(apr_palloc(pFuncPool, sizeof(*new_entry)), 0, sizeof(*new_entry));

	new_entry->name = apr_pstrdup (pFuncPool, stFolder);
	new_entry->revision = -1;
	new_entry->kind = svn_node_dir;
	new_entry->schedule = svn_wc_schedule_normal;
	new_entry->copyfrom_rev = -1;
	new_entry->cmt_rev = -1;

	apr_hash_set(entries, stFolder, APR_HASH_KEY_STRING, new_entry);

	SVN_CALL_SVNERROR(svn_wc__entries_write(entries, dir_access, pFuncPool));

	END_SVN_CALL

	return hr;
}

STDMETHODIMP CSVNClient::GetConflictRev(BSTR sPath, BSTR *sOld, BSTR *sNew, BSTR *sWrk)
{
	BEGIN_SVN_CALL

	CString stPath = Ole2Uri(sPath, pFuncPool);
	CString stFolder = stPath.Mid(stPath.ReverseFind('/')+1);
	CString stParent = stPath.Left(stPath.ReverseFind('/'));

	apr_hash_t *entries;
    svn_wc_adm_access_t *dir_access = (svn_wc_adm_access_t *)memset(apr_palloc(pFuncPool, sizeof(*dir_access)), 0, sizeof(*dir_access));
	dir_access->type = svn_wc_adm_access_t::svn_wc__adm_access_unlocked;
	dir_access->path = apr_pstrdup(pFuncPool, stParent);
	dir_access->pool = pFuncPool;

	SVN_CALL_SVNERROR(svn_wc_entries_read(&entries, dir_access, FALSE, pFuncPool));

    svn_wc_entry_t *entry = 0;
	entry = (svn_wc_entry_t *)apr_hash_get(entries, stFolder, APR_HASH_KEY_STRING);

	if (entry)
	{
		CComBSTR bsrtTemp;
		bsrtTemp = Utf82Ole(entry->conflict_old);
		bsrtTemp.CopyTo(sOld);
		bsrtTemp.Empty();

		bsrtTemp = Utf82Ole(entry->conflict_new);
		bsrtTemp.CopyTo(sNew);
		bsrtTemp.Empty();

		bsrtTemp = Utf82Ole(entry->conflict_wrk);
		bsrtTemp.CopyTo(sWrk);
	}

	END_SVN_CALL

	return hr;
}

STDMETHODIMP CSVNClient::GetSvnUrl(BSTR sPath, BSTR *sUrl)
{
	TRACELOG_BEGIN("GetSvnUrl")
	TRACELOG_TITLE("sPath: ")	TRACELOG_WCHAR(sPath)		TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	CString stPath = Ole2Utf8(sPath);
	if (stPath.GetAt(stPath.GetLength()-1)=='/')
		stPath = stPath.Left(stPath.GetLength()-1);

	CString stName = stPath.Mid(stPath.ReverseFind('/')+1);
	CString stParent = stPath.Left(stPath.ReverseFind('/'));

	apr_hash_t *entries;
    svn_wc_adm_access_t *dir_access = (svn_wc_adm_access_t *)memset(apr_palloc(pFuncPool, sizeof(*dir_access)), 0, sizeof(*dir_access));
	dir_access->type = svn_wc_adm_access_t::svn_wc__adm_access_write_lock;
	dir_access->pool = pFuncPool;

    svn_wc_entry_t *entry = 0;
	if ((GetFileAttributes(stPath) & FILE_ATTRIBUTE_DIRECTORY) > 0)
	{
		dir_access->path = apr_pstrdup(pFuncPool, stPath);
		SVN_CALL_SVNERROR(svn_wc_entries_read(&entries, dir_access, FALSE, pFuncPool));

		entry = (svn_wc_entry_t *)apr_hash_get(entries, "", APR_HASH_KEY_STRING);
	}
	else
	{
		dir_access->path = apr_pstrdup(pFuncPool, stParent);
		SVN_CALL_SVNERROR(svn_wc_entries_read(&entries, dir_access, TRUE, pFuncPool));

		entry = (svn_wc_entry_t *)apr_hash_get(entries, stName, APR_HASH_KEY_STRING);
	}

	if (entry)
	{
		CComBSTR bsrtTemp;
		bsrtTemp = Uri2Ole(entry->url, pFuncPool);
		bsrtTemp.CopyTo(sUrl);

 		TRACELOG_TITLE("sUrl: ")	TRACELOG_WCHAR(*sUrl)		TRACELOG_NEWLINE
	}



	END_SVN_CALL

	TRACELOG_END("GetSvnUrl")

	return hr;
}

STDMETHODIMP CSVNClient::SvnTimeToHumanString(CY cyTime, BSTR *sTime)
{
	BEGIN_SVN_CALL
		
		CComBSTR bsrtTemp;
		bsrtTemp = Utf82Ole(svn_time_to_human_cstring(cyTime.int64, pFuncPool));
		bsrtTemp.CopyTo(sTime);

		// Dummy macro to suppress "unreferenced label" warning
		SVN_CALL_SVNERROR(NULL);

	END_SVN_CALL

	return hr;
}

STDMETHODIMP CSVNClient::Switch(BSTR bstrPath, BSTR bstrSwitchURL, BSTR bstrRevision, VARIANT_BOOL bRecursive)
{
	TRACELOG_BEGIN("Switch")

	TRACELOG_TITLE("sPath: ")					TRACELOG_WCHAR(bstrPath)					TRACELOG_NEWLINE
	TRACELOG_TITLE("sSwitchURL: ")				TRACELOG_WCHAR(bstrSwitchURL)				TRACELOG_NEWLINE
	TRACELOG_TITLE("sRevision: ")				TRACELOG_WCHAR(bstrRevision)				TRACELOG_NEWLINE
	TRACELOG_TITLE("bRecursive: ")				TRACELOG_NUMBER(bRecursive)					TRACELOG_NEWLINE

	BEGIN_SVN_CALL

	CString sPath = Ole2Uri(bstrPath, pFuncPool);
	CString sSwitchURL = Ole2Uri(bstrSwitchURL, pFuncPool);

	svn_opt_revision_t revision;
	string_to_revision(revision, Ole2Utf8(bstrRevision));

	SVN_CALL_SVNERROR(svn_client_switch(NULL, sPath, sSwitchURL, &revision, bRecursive == VARIANT_TRUE,
			m_pCtx, pFuncPool));

	END_SVN_CALL

	TRACELOG_END("Switch")
	TRACELOG_TITLE("return: ")	TRACELOG_NUMBER(hr)	TRACELOG_NEWLINE

	return hr;
}
